from langchain_community.tools import BraveSearch
from typing import Dict, List, Any, Optional
import os
from langchain_core.language_models import BaseLLM

from local_deep_research.web_search_engines.search_engine_base import BaseSearchEngine
from local_deep_research import config
import logging
logger = logging.getLogger(__name__)

class BraveSearchEngine(BaseSearchEngine):
    """Brave search engine implementation with two-phase approach"""

    def __init__(self,
                max_results: int = 10,
                region: str = "US",
                time_period: str = "y",
                safe_search: bool = True,
                search_language: str = "English",
                api_key: Optional[str] = None,
                language_code_mapping: Optional[Dict[str, str]] = None,
                llm: Optional[BaseLLM] = None,
                include_full_content: bool = True,
                max_filtered_results: Optional[int] = None,
                **kwargs):
        """
        Initialize the Brave search engine.
        
        Args:
            max_results: Maximum number of search results
            region: Region code for search results
            time_period: Time period for search results
            safe_search: Whether to enable safe search
            search_language: Language for search results
            api_key: Brave Search API key (can also be set in BRAVE_API_KEY env)
            language_code_mapping: Mapping from language names to codes
            llm: Language model for relevance filtering
            include_full_content: Whether to include full webpage content in results
            max_filtered_results: Maximum number of results to keep after filtering
            **kwargs: Additional parameters (ignored but accepted for compatibility)
        """
        # Initialize the BaseSearchEngine with LLM, max_filtered_results, and max_results
        super().__init__(llm=llm, max_filtered_results=max_filtered_results, max_results=max_results)
        self.include_full_content = include_full_content
        
        # Set up language code mapping
        if language_code_mapping is None:
            language_code_mapping = {
                "english": "en",
                "spanish": "es",
                "chinese": "zh",
                "hindi": "hi",
                "french": "fr",
                "arabic": "ar",
                "bengali": "bn",
                "portuguese": "pt",
                "russian": "ru",
            }
        
        # Get API key
        brave_api_key = api_key or os.getenv("BRAVE_API_KEY")
        if not brave_api_key:
            raise ValueError("BRAVE_API_KEY not found. Please provide api_key or set the BRAVE_API_KEY environment variable.")
        
        # Get language code
        language_code = language_code_mapping.get(search_language.lower(), "en")
        
        # Convert time period format to Brave's format
        brave_time_period = f"p{time_period}"
        
        # Convert safe search to Brave's format
        brave_safe_search = "moderate" if safe_search else "off"
        
        # Initialize Brave Search
        self.engine = BraveSearch.from_api_key(
            api_key=brave_api_key,
            search_kwargs={
                "count": min(20, max_results),
                "country": region.upper(),
                "search_lang": language_code,
                "safesearch": brave_safe_search,
                "freshness": brave_time_period,
            }
        )
        
        # Set user agent for Brave Search
        os.environ["USER_AGENT"] = "Local Deep Research/1.0"
        
        # If full content is requested, initialize FullSearchResults
        if include_full_content:
            # Import FullSearchResults only if needed
            try:
                from local_deep_research.web_search_engines.engines.full_search import FullSearchResults
                self.full_search = FullSearchResults(
                    llm=llm,
                    web_search=self.engine,
                    language=search_language,
                    max_results=max_results,
                    region=region,
                    time=time_period,
                    safesearch=brave_safe_search
                )
            except ImportError:
                logger.warning("Warning: FullSearchResults not available. Full content retrieval disabled.")
                self.include_full_content = False
    
    def _get_previews(self, query: str) -> List[Dict[str, Any]]:
        """
        Get preview information from Brave Search.
        
        Args:
            query: The search query
            
        Returns:
            List of preview dictionaries
        """
        logger.info("Getting search results from Brave Search")
        
        try:
            # Get search results from Brave Search
            raw_results = self.engine.run(query[:400])
            
            # Parse results if they're in string format
            if isinstance(raw_results, str):
                try:
                    import json
                    raw_results = json.loads(raw_results)
                except json.JSONDecodeError:
                    logger.error("Error: Unable to parse BraveSearch response as JSON.")
                    return []
            
            # Format results as previews
            previews = []
            for i, result in enumerate(raw_results):
                preview = {
                    "id": i,  # Use index as ID
                    "title": result.get("title", ""),
                    "link": result.get("link", ""),
                    "snippet": result.get("snippet", ""),
                    "displayed_link": result.get("link", ""),
                    "position": i
                }
                
                # Store full Brave result for later
                preview["_full_result"] = result
                
                previews.append(preview)
            
            # Store the previews for potential full content retrieval
            self._search_results = previews
            
            return previews
            
        except Exception as e:
            logger.error(f"Error getting Brave Search results: {e}")
            return []
    
    def _get_full_content(self, relevant_items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Get full content for the relevant search results.
        If include_full_content is True and FullSearchResults is available,
        retrieves full webpage content for the results.
        
        Args:
            relevant_items: List of relevant preview dictionaries
            
        Returns:
            List of result dictionaries with full content if requested
        """
        # Check if we should get full content
        if hasattr(config, 'SEARCH_SNIPPETS_ONLY') and config.SEARCH_SNIPPETS_ONLY:
            logger.info("Snippet-only mode, skipping full content retrieval")
            
            # Return the relevant items with their full Brave information
            results = []
            for item in relevant_items:
                # Use the full result if available, otherwise use the preview
                if "_full_result" in item:
                    result = item["_full_result"]
                    # Remove temporary field
                    if "_full_result" in result:
                        del result["_full_result"]
                else:
                    result = item
                    
                results.append(result)
                
            return results
            
        # If full content retrieval is enabled
        if self.include_full_content and hasattr(self, 'full_search'):
            logger.info("Retrieving full webpage content")
            
            try:
                # Extract only the links from relevant items
                links = [item.get("link") for item in relevant_items if item.get("link")]
                
                # Use FullSearchResults to get full content
                results_with_content = self.full_search._get_full_content(relevant_items)
                
                return results_with_content
                
            except Exception as e:
                logger.error(f"Error retrieving full content: {e}")
                # Fall back to returning the items without full content
        
        # Return items with their full Brave information
        results = []
        for item in relevant_items:
            # Use the full result if available, otherwise use the preview
            if "_full_result" in item:
                result = item["_full_result"].copy()
                # Remove temporary field
                if "_full_result" in result:
                    del result["_full_result"]
            else:
                result = item.copy()
                if "_full_result" in result:
                    del result["_full_result"]
                
            results.append(result)
            
        return results
    
    def run(self, query: str) -> List[Dict[str, Any]]:
        """
        Execute a search using Brave Search with the two-phase approach.
        
        Args:
            query: The search query
            
        Returns:
            List of search results
        """
        logger.info("---Execute a search using Brave Search---")
        
        # Use the implementation from the parent class which handles all phases
        results = super().run(query)
        
        # Clean up
        if hasattr(self, '_search_results'):
            del self._search_results
            
        return results



import requests
import logging
from typing import Dict, List, Any, Optional, Tuple
import os
from datetime import datetime, timedelta
from langchain_core.language_models import BaseLLM

from local_deep_research.web_search_engines.search_engine_base import BaseSearchEngine
from local_deep_research import config
from local_deep_research.utilties.search_utilities import remove_think_tags

# Setup logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class GuardianSearchEngine(BaseSearchEngine):
    """Enhanced Guardian API search engine implementation with LLM query optimization"""
    
    def __init__(self, 
                max_results: int = 10, 
                api_key: Optional[str] = None,
                from_date: Optional[str] = None,
                to_date: Optional[str] = None,
                section: Optional[str] = None,
                order_by: str = "relevance",
                llm: Optional[BaseLLM] = None,
                max_filtered_results: Optional[int] = None,
                optimize_queries: bool = True,
                adaptive_search: bool = True):
        """
        Initialize The Guardian search engine with enhanced features.
        
        Args:
            max_results: Maximum number of search results
            api_key: The Guardian API key (can also be set in GUARDIAN_API_KEY env)
            from_date: Start date for search (YYYY-MM-DD format, default 1 month ago)
            to_date: End date for search (YYYY-MM-DD format, default today)
            section: Filter by section (e.g., "politics", "technology", "sport")
            order_by: Sort order ("relevance", "newest", "oldest")
            llm: Language model for relevance filtering and query optimization
            max_filtered_results: Maximum number of results to keep after filtering
            optimize_queries: Whether to optimize queries using LLM
            adaptive_search: Whether to use adaptive search (adjusting date ranges)
        """
        # Initialize the BaseSearchEngine with LLM, max_filtered_results, and max_results
        super().__init__(llm=llm, max_filtered_results=max_filtered_results, max_results=max_results)
        self.api_key = api_key or os.getenv("GUARDIAN_API_KEY")
        self.optimize_queries = optimize_queries
        self.adaptive_search = adaptive_search
        
        if not self.api_key:
            raise ValueError("Guardian API key not found. Please provide api_key or set the GUARDIAN_API_KEY environment variable.")
        
        # Set date ranges if not provided
        if not from_date:
            # Default to one month ago
            one_month_ago = datetime.now() - timedelta(days=30)
            self.from_date = one_month_ago.strftime("%Y-%m-%d")
        else:
            self.from_date = from_date
            
        if not to_date:
            # Default to today
            self.to_date = datetime.now().strftime("%Y-%m-%d")
        else:
            self.to_date = to_date
            
        self.section = section
        self.order_by = order_by
        self._original_date_params = {
            "from_date": self.from_date,
            "to_date": self.to_date
        }
        
        # API base URL
        self.api_url = "https://content.guardianapis.com/search"
    
    def _optimize_query_for_guardian(self, query: str) -> str:
        """
        Optimize a natural language query for Guardian search.
        Uses LLM to transform questions into effective news search queries.
        
        Args:
            query: Natural language query
            
        Returns:
            Optimized query string for Guardian
        """
        # Handle extremely long queries by truncating first
        if len(query) > 150:
            simple_query = " ".join(query.split()[:10])
            logger.info(f"Query too long ({len(query)} chars), truncating to: {simple_query}")
            query = simple_query
        
        if not self.llm or not self.optimize_queries:
            # Return original query if no LLM available or optimization disabled
            return query
            
        try:
            # Prompt for query optimization
            prompt = f"""Transform this natural language question into a very short Guardian news search query.

Original query: "{query}"

CRITICAL RULES:
1. ONLY RETURN THE EXACT SEARCH QUERY - NO EXPLANATIONS, NO COMMENTS
2. Keep it EXTREMELY BRIEF - MAXIMUM 3-4 words total
3. Focus only on the main topic/person/event
4. Include proper names when relevant
5. Remove ALL unnecessary words
6. DO NOT use Boolean operators (no AND/OR)
7. DO NOT use quotes

EXAMPLE CONVERSIONS:
✓ "What's the impact of rising interest rates on UK housing market?" → "UK housing rates"
✓ "Latest developments in the Ukraine-Russia peace negotiations" → "Ukraine Russia negotiations"
✓ "How are tech companies responding to AI regulation?" → "tech AI regulation"
✓ "What is Donald Trump's current political activity?" → "Trump political activity"

Return ONLY the extremely brief search query.
"""
            
            # Get response from LLM
            response = self.llm.invoke(prompt)
            optimized_query = remove_think_tags(response.content).strip()
            
            # Clean up the query - remove any explanations
            lines = optimized_query.split('\n')
            for line in lines:
                line = line.strip()
                if line and not line.lower().startswith(('here', 'i would', 'the best', 'this query')):
                    optimized_query = line
                    break
            
            # Remove any quotes that wrap the entire query
            if optimized_query.startswith('"') and optimized_query.endswith('"') and optimized_query.count('"') == 2:
                optimized_query = optimized_query[1:-1]
            
            logger.info(f"Original query: '{query}'")
            logger.info(f"Optimized for Guardian: '{optimized_query}'")
            
            return optimized_query
            
        except Exception as e:
            logger.error(f"Error optimizing query: {e}")
            return query  # Fall back to original query on error
    
    def _adapt_dates_for_query_type(self, query: str) -> None:
        """
        Adapt date range based on query type (historical vs current).
        
        Args:
            query: The search query
        """
        # Fast path - for very short queries, default to recent news
        if len(query.split()) <= 4:
            logger.info("Short query detected, defaulting to recent news")
            # Default to 60 days for short queries
            recent = (datetime.now() - timedelta(days=60)).strftime("%Y-%m-%d")
            self.from_date = recent
            self.order_by = "newest"
            return
        
        if not self.llm or not self.adaptive_search:
            return
            
        try:
            prompt = f"""Is this query asking about HISTORICAL events or CURRENT events?

Query: "{query}"

ONE WORD ANSWER ONLY:
- "HISTORICAL" if about past events (older than 1 year)
- "CURRENT" if about recent events (within past year)
- "UNCLEAR" if can't determine

ONE WORD ONLY:"""
            
            response = self.llm.invoke(prompt)
            answer = remove_think_tags(response.content).strip().upper()
            
            # Reset to original parameters first
            self.from_date = self._original_date_params["from_date"]
            self.to_date = self._original_date_params["to_date"]
            
            if "HISTORICAL" in answer:
                # For historical queries, go back 10 years
                logger.info("Query classified as HISTORICAL - extending search timeframe")
                ten_years_ago = (datetime.now() - timedelta(days=3650)).strftime("%Y-%m-%d")
                self.from_date = ten_years_ago
                
            elif "CURRENT" in answer:
                # For current events, focus on recent content
                logger.info("Query classified as CURRENT - focusing on recent content")
                recent = (datetime.now() - timedelta(days=60)).strftime("%Y-%m-%d")
                self.from_date = recent
                self.order_by = "newest"  # Prioritize newest for current events
                
        except Exception as e:
            logger.error(f"Error adapting dates for query type: {e}")
            # Keep original date parameters on error
    
    def _adaptive_search(self, query: str) -> Tuple[List[Dict[str, Any]], str]:
        """
        Perform adaptive search that progressively adjusts parameters based on results.
        
        Args:
            query: The search query
            
        Returns:
            Tuple of (list of articles, search strategy used)
        """
        # Try with current parameters
        articles = self._get_all_data(query)
        strategy = "initial"
        
        # If no results or too few, try different strategies
        if len(articles) < 3 and self.adaptive_search:
            logger.info(f"Initial search found only {len(articles)} results, trying alternative strategies")
            
            # Try with expanded date range
            original_from_date = self.from_date
            original_order_by = self.order_by
            
            # Strategy 1: Expand to 6 months
            logger.info("Strategy 1: Expanding time range to 6 months")
            six_months_ago = (datetime.now() - timedelta(days=180)).strftime("%Y-%m-%d")
            self.from_date = six_months_ago
            
            articles1 = self._get_all_data(query)
            if len(articles1) > len(articles):
                articles = articles1
                strategy = "expanded_6mo"
                
            # Strategy 2: Expand to all time and try relevance order
            if len(articles) < 3:
                logger.info("Strategy 2: Expanding to all time with relevance ordering")
                self.from_date = "2000-01-01"  # Effectively "all time"
                self.order_by = "relevance"
                
                articles2 = self._get_all_data(query)
                if len(articles2) > len(articles):
                    articles = articles2
                    strategy = "all_time_relevance"
            
            # Strategy 3: Try removing section constraints
            if len(articles) < 3 and self.section:
                logger.info("Strategy 3: Removing section constraint")
                original_section = self.section
                self.section = None
                
                articles3 = self._get_all_data(query)
                if len(articles3) > len(articles):
                    articles = articles3
                    strategy = "no_section"
                    
                # Restore section setting
                self.section = original_section
            
            # Restore original settings
            self.from_date = original_from_date
            self.order_by = original_order_by
            
        logger.info(f"Adaptive search using strategy '{strategy}' found {len(articles)} results")
        return articles, strategy
    
    def _get_all_data(self, query: str) -> List[Dict[str, Any]]:
        """
        Get all article data from The Guardian API in a single call.
        Always requests all fields for simplicity.
        
        Args:
            query: The search query
            
        Returns:
            List of articles with all data
        """
        try:
            # Ensure query is not empty
            if not query or query.strip() == "":
                query = "news"
                logger.warning("Empty query provided, using 'news' as default")
            
            # Ensure query is not too long for API
            if len(query) > 100:
                logger.warning(f"Query too long for Guardian API ({len(query)} chars), truncating")
                query = query[:100]
            
            # Always request all fields for simplicity
            # Ensure max_results is an integer to avoid comparison errors
            page_size = min(int(self.max_results) if self.max_results is not None else 10, 50)
            
            # Log full parameters for debugging
            logger.info(f"Guardian API search query: '{query}'")
            logger.info(f"Guardian API date range: {self.from_date} to {self.to_date}")
            
            params = {
                "q": query,
                "api-key": self.api_key,
                "from-date": self.from_date,
                "to-date": self.to_date,
                "order-by": self.order_by,
                "page-size": page_size,  # API maximum is 50
                "show-fields": "headline,trailText,byline,body,publication",
                "show-tags": "keyword"
            }
            
            # Add section filter if specified
            if self.section:
                params["section"] = self.section
            
            # Log the complete request parameters (except API key)
            log_params = params.copy()
            log_params["api-key"] = "REDACTED"
            logger.info(f"Guardian API request parameters: {log_params}")
            
            # Execute the API request
            response = requests.get(self.api_url, params=params)
            response.raise_for_status()
            
            data = response.json()
            
            # Extract results from the response
            articles = data.get("response", {}).get("results", [])
            logger.info(f"Guardian API returned {len(articles)} articles")
            
            # Format results to include all data
            formatted_articles = []
            for i, article in enumerate(articles):
                if i >= self.max_results:
                    break
                    
                fields = article.get("fields", {})
                
                # Format the article with all fields
                result = {
                    "id": article.get("id", ""),
                    "title": fields.get("headline", article.get("webTitle", "")),
                    "link": article.get("webUrl", ""),
                    "snippet": fields.get("trailText", ""),
                    "publication_date": article.get("webPublicationDate", ""),
                    "section": article.get("sectionName", ""),
                    "author": fields.get("byline", ""),
                    "content": fields.get("body", ""),
                    "full_content": fields.get("body", "")
                }
                
                # Extract tags/keywords
                tags = article.get("tags", [])
                result["keywords"] = [tag.get("webTitle", "") for tag in tags if tag.get("type") == "keyword"]
                
                formatted_articles.append(result)
            
            return formatted_articles
            
        except Exception as e:
            logger.error(f"Error getting data from The Guardian API: {e}")
            return []
    
    def _get_previews(self, query: str) -> List[Dict[str, Any]]:
        """
        Get preview information for Guardian articles with enhanced optimization.
        
        Args:
            query: The search query
            
        Returns:
            List of preview dictionaries
        """
        logger.info(f"Getting articles from The Guardian API for query: {query}")
        
        # Step 1: Optimize the query using LLM
        optimized_query = self._optimize_query_for_guardian(query)
        
        # Step 2: Adapt date parameters based on query type
        self._adapt_dates_for_query_type(optimized_query)
        
        # Step 3: Perform adaptive search
        articles, strategy = self._adaptive_search(optimized_query)
        
        # Store search metadata for debugging
        self._search_metadata = {
            "original_query": query,
            "optimized_query": optimized_query,
            "strategy": strategy,
            "from_date": self.from_date,
            "to_date": self.to_date,
            "section": self.section,
            "order_by": self.order_by
        }
        
        # Store full articles for later use
        self._full_articles = {a["id"]: a for a in articles}
        
        # Return only preview fields for each article
        previews = []
        for article in articles:
            preview = {
                "id": article["id"],
                "title": article["title"],
                "link": article["link"],
                "snippet": article["snippet"],
                "publication_date": article["publication_date"],
                "section": article["section"],
                "author": article["author"],
                "keywords": article.get("keywords", [])
            }
            previews.append(preview)
        
        return previews
    
    def _get_full_content(self, relevant_items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Get full content for the relevant Guardian articles.
        Restores full content from the cached data.
        
        Args:
            relevant_items: List of relevant preview dictionaries
            
        Returns:
            List of result dictionaries with full content
        """
        logger.info(f"Adding full content to {len(relevant_items)} relevant Guardian articles")
        
        # Check if we should add full content
        if hasattr(config, 'SEARCH_SNIPPETS_ONLY') and config.SEARCH_SNIPPETS_ONLY:
            return relevant_items
            
        # Get full articles for relevant items
        results = []
        for item in relevant_items:
            article_id = item.get("id", "")
            
            # Get the full article from our cache
            if hasattr(self, '_full_articles') and article_id in self._full_articles:
                results.append(self._full_articles[article_id])
            else:
                # If not found (shouldn't happen), just use the preview
                results.append(item)
        
        return results
    
    def run(self, query: str) -> List[Dict[str, Any]]:
        """
        Execute a search using The Guardian API with the enhanced approach.
        
        Args:
            query: The search query
            
        Returns:
            List of search results
        """
        logger.info(f"---Execute a search using The Guardian (enhanced)---")
        
        # Additional safety check for None query
        if query is None:
            logger.error("None query passed to Guardian search engine")
            query = "news"
        
        try:
            # Get previews with our enhanced method
            previews = self._get_previews(query)
            
            # If no results, try one more time with a simplified query
            if not previews:
                simple_query = " ".join([w for w in query.split() if len(w) > 3][:3])
                logger.warning(f"No Guardian articles found, trying simplified query: {simple_query}")
                previews = self._get_previews(simple_query)
                
                # If still no results, try with a very generic query as last resort
                if not previews and "trump" in query.lower():
                    logger.warning("Trying last resort query: 'Donald Trump'")
                    previews = self._get_previews("Donald Trump")
                elif not previews:
                    logger.warning("Trying last resort query: 'news'")
                    previews = self._get_previews("news")
            
            # If still no results after all attempts, return empty list
            if not previews:
                logger.warning(f"No Guardian articles found after multiple attempts")
                return []
            
            # Filter for relevance if we have an LLM
            if self.llm and hasattr(self, 'max_filtered_results') and self.max_filtered_results:
                filtered_items = self._filter_for_relevance(previews, query)
                if not filtered_items:
                    # Fall back to unfiltered results if everything was filtered out
                    logger.warning("All articles filtered out, using unfiltered results")
                    filtered_items = previews[:self.max_filtered_results]
            else:
                filtered_items = previews
            
            # Get full content for relevant items
            results = self._get_full_content(filtered_items)
            
            # Add source information to make it clear these are from The Guardian
            for result in results:
                if "source" not in result:
                    result["source"] = "The Guardian"
            
            # Clean up the cache after use
            if hasattr(self, '_full_articles'):
                del self._full_articles
                
            # Restore original date parameters
            self.from_date = self._original_date_params["from_date"]
            self.to_date = self._original_date_params["to_date"]
            
            # Log search metadata if available
            if hasattr(self, '_search_metadata'):
                logger.info(f"Search metadata: {self._search_metadata}")
                del self._search_metadata
                
            return results
            
        except Exception as e:
            logger.error(f"Error in Guardian search: {e}")
            
            # Restore original date parameters on error
            self.from_date = self._original_date_params["from_date"]
            self.to_date = self._original_date_params["to_date"]
            
            return []
    
    def search_by_section(self, section: str, max_results: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Search for articles in a specific section.
        
        Args:
            section: The Guardian section name (e.g., "politics", "technology")
            max_results: Maximum number of results (defaults to self.max_results)
            
        Returns:
            List of articles in the section
        """
        original_section = self.section
        original_max_results = self.max_results
        
        try:
            # Set section and max_results for this search
            self.section = section
            if max_results:
                self.max_results = max_results
                
            # Use empty query to get all articles in the section
            return self.run("")
            
        finally:
            # Restore original values
            self.section = original_section
            self.max_results = original_max_results
    
    def get_recent_articles(self, days: int = 7, max_results: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Get recent articles from The Guardian.
        
        Args:
            days: Number of days to look back
            max_results: Maximum number of results (defaults to self.max_results)
            
        Returns:
            List of recent articles
        """
        original_from_date = self.from_date
        original_order_by = self.order_by
        original_max_results = self.max_results
        
        try:
            # Set parameters for this search
            self.from_date = (datetime.now() - timedelta(days=days)).strftime("%Y-%m-%d")
            self.order_by = "newest"
            if max_results:
                self.max_results = max_results
                
            # Use empty query to get all recent articles
            return self.run("")
            
        finally:
            # Restore original values
            self.from_date = original_from_date
            self.order_by = original_order_by
            self.max_results = original_max_results


"""
LLM configuration for Local Deep Research.

This file controls which language models are available and how they're configured.
You can customize model selection, parameters, and fallbacks here.
"""

from langchain_anthropic import ChatAnthropic
from langchain_openai import ChatOpenAI
from langchain_ollama import ChatOllama
from langchain_community.llms import VLLM
from local_deep_research.config import settings
import os
import logging

# Initialize environment
logger = logging.getLogger(__name__)

# Valid provider options
VALID_PROVIDERS = ["ollama", "openai", "anthropic", "vllm", "openai_endpoint", "lmstudio", "llamacpp", "none"]

# ================================
# LLM FUNCTIONS
# ================================

def get_llm(model_name=None, temperature=None, provider=None):
    """
    Get LLM instance based on model name and provider.
    
    Args:
        model_name: Name of the model to use (if None, uses settings.llm.model)
        temperature: Model temperature (if None, uses settings.llm.temperature)
        provider: Provider to use (if None, uses settings.llm.provider)
    
    Returns:
        A LangChain LLM instance
    """
    # Use settings values for parameters if not provided
    if model_name is None:
        model_name = settings.llm.model
    
    if temperature is None:
        temperature = settings.llm.temperature
        
    if provider is None:
        provider = settings.llm.provider.lower()
        if provider not in VALID_PROVIDERS:
            logger.error(f"Invalid provider in settings: {provider}")
            raise ValueError(f"Invalid provider: {provider}. Must be one of: {VALID_PROVIDERS}")
    
    # Common parameters for all models
    common_params = {
        "temperature": temperature,
        "max_tokens": settings.llm.max_tokens,
    }
    
    # Handle different providers
    if provider == "anthropic":
        api_key = settings.get('ANTHROPIC_API_KEY', '')
        if not api_key:
            logger.warning("ANTHROPIC_API_KEY not found. Falling back to default model.")
            return get_fallback_model(temperature)
        
        return ChatAnthropic(
            model=model_name, anthropic_api_key=api_key, **common_params
        )
    
    elif provider == "openai":
        api_key = settings.get('OPENAI_API_KEY', '')
        if not api_key:
            logger.warning("OPENAI_API_KEY not found. Falling back to default model.")
            return get_fallback_model(temperature)
        
        return ChatOpenAI(model=model_name, api_key=api_key, **common_params)
    
    elif provider == "openai_endpoint":
        api_key = settings.get('OPENAI_ENDPOINT_API_KEY', '')

        if not api_key:
            logger.warning("OPENAI_ENDPOINT_API_KEY not found. Falling back to default model.")
            return get_fallback_model(temperature)
        
        # Get endpoint URL from settings
        openai_endpoint_url = settings.llm.openai_endpoint_url
        
        return ChatOpenAI(
            model=model_name, 
            api_key=api_key,
            openai_api_base=openai_endpoint_url, 
            **common_params
        )
    
    elif provider == "vllm":
        try:
            return VLLM(
                model=model_name,
                trust_remote_code=True,
                max_new_tokens=128,
                top_k=10,
                top_p=0.95,
                temperature=temperature,
            )
        except Exception as e:
            logger.error(f"Error loading VLLM model: {e}")
            logger.warning("Falling back.")
            return get_fallback_model(temperature)
    
    elif provider == "ollama":
        try:
            # Use the configurable Ollama base URL
            base_url = settings.get('OLLAMA_BASE_URL', settings.llm.get('ollama_base_url', 'http://localhost:11434'))
            return ChatOllama(model=model_name, base_url=base_url, **common_params)
        except Exception as e:
            logger.error(f"Error loading Ollama model: {e}")
            return get_fallback_model(temperature)
    
    elif provider == "lmstudio":

            # LM Studio supports OpenAI API format, so we can use ChatOpenAI directly
            lmstudio_url = settings.llm.get('lmstudio_url', "http://localhost:1234")
            
            return ChatOpenAI(
                model=model_name,
                api_key="lm-studio",  # LM Studio doesn't require a real API key
                base_url=f"{lmstudio_url}/v1",  # Use the configured URL with /v1 endpoint
                temperature=temperature,
                max_tokens=settings.llm.max_tokens
            )

 
    elif provider == "llamacpp":

            # Import LlamaCpp
            from langchain_community.llms import LlamaCpp
            
            # Get LlamaCpp model path from settings
            model_path = settings.llm.get('llamacpp_model_path', "")
            if not model_path:
                logger.error("llamacpp_model_path not set in settings")
                raise ValueError("llamacpp_model_path not set in settings.toml")
                
            # Get additional LlamaCpp parameters
            n_gpu_layers = settings.llm.get('llamacpp_n_gpu_layers', 1)
            n_batch = settings.llm.get('llamacpp_n_batch', 512)
            f16_kv = settings.llm.get('llamacpp_f16_kv', True)
            
            # Create LlamaCpp instance
            return LlamaCpp(
                model_path=model_path,
                temperature=temperature,
                max_tokens=settings.llm.max_tokens,
                n_gpu_layers=n_gpu_layers,
                n_batch=n_batch,
                f16_kv=f16_kv,
                verbose=True
            )
    
    else:
        return get_fallback_model(temperature)

def get_fallback_model(temperature=None):
    """Create a dummy model for when no providers are available"""
    from langchain_community.llms.fake import FakeListLLM
    return FakeListLLM(
        responses=["No language models are available. Please install Ollama or set up API keys."]
    )

# ================================
# COMPATIBILITY FUNCTIONS
# ================================

def get_available_provider_types():
    """Return available model providers"""
    providers = {}
    
    if is_ollama_available():
        providers["ollama"] = "Ollama (local models)"
    
    if is_openai_available():
        providers["openai"] = "OpenAI API"
    
    if is_anthropic_available():
        providers["anthropic"] = "Anthropic API"
    
    if is_openai_endpoint_available():
        providers["openai_endpoint"] = "OpenAI-compatible Endpoint"
    
    if is_lmstudio_available():
        providers["lmstudio"] = "LM Studio (local models)"
        
    if is_llamacpp_available():
        providers["llamacpp"] = "LlamaCpp (local models)"
    
    # Check for VLLM capability
    try:
        import torch
        import transformers
        providers["vllm"] = "VLLM (local models)"
    except ImportError:
        pass
    
    # Default fallback
    if not providers:
        providers["none"] = "No model providers available"
    
    return providers

# ================================
# HELPER FUNCTIONS
# ================================

def is_openai_available():
    """Check if OpenAI is available"""
    try:
        api_key = settings.get('OPENAI_API_KEY', '')
        return bool(api_key)
    except:
        return False

def is_anthropic_available():
    """Check if Anthropic is available"""
    try:
        api_key = settings.get('ANTHROPIC_API_KEY', '')
        return bool(api_key)
    except:
        return False

def is_openai_endpoint_available():
    """Check if OpenAI endpoint is available"""
    try:
        api_key = settings.get('OPENAI_ENDPOINT_API_KEY', '')
        return bool(api_key) 
    except:
        return False

def is_ollama_available():
    """Check if Ollama is running"""
    try:
        import requests
        base_url = settings.get('OLLAMA_BASE_URL', settings.llm.get('ollama_base_url', 'http://localhost:11434'))
        response = requests.get(f"{base_url}/api/tags", timeout=1.0)
        return response.status_code == 200
    except:
        return False

def is_vllm_available():
    """Check if VLLM capability is available"""
    try:
        import torch
        import transformers
        return True
    except ImportError:
        return False

def is_lmstudio_available():
    """Check if LM Studio is available"""
    try:
        import requests
        lmstudio_url = settings.llm.get('lmstudio_url', 'http://localhost:1234')
        # LM Studio typically uses OpenAI-compatible endpoints
        response = requests.get(f"{lmstudio_url}/v1/models", timeout=1.0)
        return response.status_code == 200
    except:
        return False

def is_llamacpp_available():
    """Check if LlamaCpp is available and configured"""
    try:
        from langchain_community.llms import LlamaCpp
        model_path = settings.llm.get('llamacpp_model_path', '')
        return bool(model_path) and os.path.exists(model_path)
    except:
        return False

def get_available_providers():
    """Get dictionary of available providers"""
    return get_available_provider_types()

# Log which providers are available
AVAILABLE_PROVIDERS = get_available_providers()
logger.info(f"Available providers: {list(AVAILABLE_PROVIDERS.keys())}")

# Check if selected provider is available
selected_provider = settings.llm.provider.lower()
if selected_provider not in AVAILABLE_PROVIDERS and selected_provider != "none":
    logger.warning(f"Selected provider {selected_provider} is not available.")


from .search_system import AdvancedSearchSystem
from typing import Dict
from .config import settings

def print_report(report: Dict):
    """Print and save the report in a readable format"""

    # Print to console in readable format
    print("\n=== GENERATED REPORT ===\n")

    # Print content
    print(report["content"])



    # Save to file in markdown format
    with open("report.md", "w", encoding="utf-8") as markdown_file:
        # Write content
        markdown_file.write(report["content"])

        # Write metadata at the end of the file
        markdown_file.write("\n\n---\n\n")
        markdown_file.write("## Report Metadata\n")

        markdown_file.write(f"- Query: {report['metadata']['query']}\n")

    print(f"\nReport has been saved to report.md")


from .report_generator import IntegratedReportGenerator

report_generator = IntegratedReportGenerator()



def main():
    import os
    import logging
    from .utilties.setup_utils import setup_user_directories
    
    # Configure logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)
    logger.info(f"Starting with settings: iterations={settings.search.iterations}, "
                f"questions_per_iteration={settings.search.questions_per_iteration}")
        
    # Explicitly run setup
    logger.info("Initializing configuration...")
    setup_user_directories()
    
    system = AdvancedSearchSystem()

    print("Welcome to the Advanced Research System")
    print("Type 'quit' to exit")

    while True:
        print("\nSelect output type:")
        print("1) Quick Summary (Generated in a few minutes)")
        print(
            "2) Detailed Research Report (Recommended for deeper analysis - may take several hours)"
        )
        choice = input("Enter number (1 or 2): ").strip()

        while choice not in ["1", "2"]:
            print("\nInvalid input. Please enter 1 or 2:")
            print("1) Quick Summary (Generated in a few minutes)")
            print(
                "2) Detailed Research Report (Recommended for deeper analysis - may take several hours)"
            )
            choice = input("Enter number (1 or 2): ").strip()

        query = input("\nEnter your research query: ").strip()

        if query.lower() == "quit":
            break

        # System will automatically use updated configuration
        # through the automatic reloading in get_llm() and get_search()

        if choice == "1":
            print("\nResearching... This may take a few minutes.\n")
        else:
            print(
                "\nGenerating detailed report... This may take several hours. Please be patient as this enables deeper analysis.\n"
            )

        results = system.analyze_topic(query)
        if results:
            if choice == "1":
                # Quick Summary
                print("\n=== QUICK SUMMARY ===")
                if results["findings"] and len(results["findings"]) > 0:
                    initial_analysis = [
                        finding["content"] for finding in results["findings"]
                    ]
                    print(initial_analysis)

            else:
                # Full Report
                final_report = report_generator.generate_report(
                    results, query
                )
                print("\n=== RESEARCH REPORT ===")
                print_report(final_report)

                print("\n=== RESEARCH METRICS ===")
                print(f"Search Iterations: {results['iterations']}")

        else:
            print("Research failed. Please try again.")

if __name__ == "__main__":
    main()


import logging
import os
from typing import Dict, List, Any, Optional

from local_deep_research.web_search_engines.search_engine_base import BaseSearchEngine
from local_deep_research.web_search_engines.search_engines_config import SEARCH_ENGINES
from local_deep_research.web_search_engines.search_engine_factory import create_search_engine
from local_deep_research.web_search_engines.engines.search_engine_wikipedia import WikipediaSearchEngine
from local_deep_research import config

# Setup logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class MetaSearchEngine(BaseSearchEngine):
    """
    LLM-powered meta search engine that intelligently selects and uses
    the appropriate search engines based on query analysis
    """
    
    def __init__(self, 
                 llm,
                 max_results: int = 10,
                 use_api_key_services: bool = True,
                 max_engines_to_try: int = 3,
                 max_filtered_results: Optional[int] = None,
                 **kwargs):
        """
        Initialize the meta search engine.
        
        Args:
            llm: Language model instance for query classification and relevance filtering
            max_results: Maximum number of search results to return
            use_api_key_services: Whether to include services that require API keys
            max_engines_to_try: Maximum number of engines to try before giving up
            max_filtered_results: Maximum number of results to keep after filtering
            **kwargs: Additional parameters (ignored but accepted for compatibility)
        """
        # Initialize the BaseSearchEngine with the LLM and max_filtered_results
        super().__init__(llm=llm, max_filtered_results=max_filtered_results)
        
        self.max_results = max_results
        self.use_api_key_services = use_api_key_services
        self.max_engines_to_try = max_engines_to_try
        
        # Cache for engine instances
        self.engine_cache = {}
        
        # Get available engines (excluding 'meta' and 'auto')
        self.available_engines = self._get_available_engines()
        logger.info(f"Meta Search Engine initialized with {len(self.available_engines)} available engines: {', '.join(self.available_engines)}")
        
        # Create a fallback engine in case everything else fails
        self.fallback_engine = WikipediaSearchEngine(
            max_results=max_results, 
            llm=llm,
            max_filtered_results=max_filtered_results
        )
    
    def _get_available_engines(self) -> List[str]:
        """Get list of available engines, excluding 'meta' and 'auto'"""
        # Filter out 'meta' and 'auto' and check API key availability
        available = []
        for name, config in SEARCH_ENGINES.items():
            if name in ["meta", "auto"]:
                continue
                
            if config.get("requires_api_key", False) and not self.use_api_key_services:
                continue
                
            if config.get("requires_api_key", False):
                api_key_env = config.get("api_key_env")
                api_key = os.getenv(api_key_env) if api_key_env else None
                if not api_key:
                    continue
                    
            available.append(name)
        
        # Make sure we have at least one engine available
        if not available and "wikipedia" in SEARCH_ENGINES:
            available.append("wikipedia")
            
        return available
    
    def analyze_query(self, query: str) -> List[str]:
        """
        Use the LLM to analyze the query and return a ranked list of
        recommended search engines to try
        """
        if not self.available_engines:
            logger.warning("No search engines available")
            return []
        engine_descriptions = []
        for name in self.available_engines:
            logger.info(f"Processing search engine: {name}")
            try:
                description = f"- {name.upper()}: Good for {', '.join(SEARCH_ENGINES[name]['strengths'][:3])}. " \
                            f"Weaknesses: {', '.join(SEARCH_ENGINES[name]['weaknesses'][:2])}. " \
                            f"Reliability: {SEARCH_ENGINES[name]['reliability']*100:.0f}%"
                engine_descriptions.append(description)
            except KeyError as e:
                logger.error(f"Missing key for engine {name}: {e}")
                # Add a basic description for engines with missing configuration
                engine_descriptions.append(f"- {name.upper()}: General purpose search engine.")
            except Exception as e:
                logger.error(f"Error processing engine {name}: {e}")
                engine_descriptions.append(f"- {name.upper()}: General purpose search engine.")

        engine_descriptions = "\n".join(engine_descriptions)
        
        prompt = f"""Analyze this search query and rank the available search engines in order of most to least appropriate for answering it.
        
Query: "{query}"

Available search engines:
{engine_descriptions}

Consider:
1. The nature of the query (factual, academic, product-related, news, etc.)
2. The strengths and weaknesses of each engine
3. The reliability of each engine

Return ONLY a comma-separated list of search engine names in your recommended order. Example: "wikipedia,arxiv,duckduckgo"
Do not include any engines that are not listed above. Only return the comma-separated list, nothing else."""

        # Get response from LLM
        try:
            response = self.llm.invoke(prompt)
            content = response.content.strip()
            
            # Parse the response into a list of engine names
            engine_names = [name.strip().lower() for name in content.split(',')]
            
            # Filter out any invalid engine names
            valid_engines = [name for name in engine_names if name in self.available_engines]
            
            # If no valid engines were returned, use default order based on reliability
            if not valid_engines:
                valid_engines = sorted(
                    self.available_engines, 
                    key=lambda x: SEARCH_ENGINES[x]["reliability"],
                    reverse=True
                )
            
            return valid_engines
        except Exception as e:
            logger.error(f"Error analyzing query with LLM: {str(e)}")
            # Fall back to reliability-based ordering
            return sorted(
                self.available_engines, 
                key=lambda x: SEARCH_ENGINES[x]["reliability"],
                reverse=True
            )
    
    def _get_previews(self, query: str) -> List[Dict[str, Any]]:
        """
        Get preview information by selecting the best search engine for this query.
        
        Args:
            query: The search query
            
        Returns:
            List of preview dictionaries
        """
        # Get ranked list of engines for this query
        ranked_engines = self.analyze_query(query)
        
        if not ranked_engines:
            logger.warning("No suitable search engines found for query, using fallback engine")
            return self.fallback_engine._get_previews(query)
        
        # Limit the number of engines to try
        engines_to_try = ranked_engines[:self.max_engines_to_try]
        
        logger.info(f"Search plan created. Will try these engines in order: {', '.join(engines_to_try)}")
        
        all_errors = []
        # Try each engine in order
        for engine_name in engines_to_try:
            logger.info(f"Trying search engine: {engine_name}")
            
            # Get or create the engine instance
            engine = self._get_engine_instance(engine_name)
            
            if not engine:
                logger.warning(f"Failed to initialize {engine_name}, skipping")
                all_errors.append(f"Failed to initialize {engine_name}")
                continue
            
            try:
                # Get previews from this engine
                previews = engine._get_previews(query)
                
                # If search was successful, return results
                if previews and len(previews) > 0:
                    logger.info(f"Successfully got {len(previews)} preview results from {engine_name}")
                    # Store selected engine for later use
                    self._selected_engine = engine
                    self._selected_engine_name = engine_name
                    return previews
                
                logger.info(f"{engine_name} returned no previews")
                all_errors.append(f"{engine_name} returned no previews")
            
            except Exception as e:
                error_msg = f"Error getting previews from {engine_name}: {str(e)}"
                logger.error(error_msg)
                all_errors.append(error_msg)
        
        # If we reach here, all engines failed, use fallback
        logger.warning(f"All engines failed or returned no preview results: {', '.join(all_errors)}")
        logger.info("Using fallback Wikipedia engine for previews")
        self._selected_engine = self.fallback_engine
        self._selected_engine_name = "wikipedia"
        return self.fallback_engine._get_previews(query)
    
    def _get_full_content(self, relevant_items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Get full content using the engine that provided the previews.
        
        Args:
            relevant_items: List of relevant preview dictionaries
            
        Returns:
            List of result dictionaries with full content
        """
        # Check if we should get full content
        if hasattr(config, 'SEARCH_SNIPPETS_ONLY') and config.SEARCH_SNIPPETS_ONLY:
            logger.info("Snippet-only mode, skipping full content retrieval")
            return relevant_items
            
        logger.info("Getting full content for relevant items")
        
        # Use the selected engine to get full content
        if hasattr(self, '_selected_engine'):
            try:
                logger.info(f"Using {self._selected_engine_name} to get full content")
                return self._selected_engine._get_full_content(relevant_items)
            except Exception as e:
                logger.error(f"Error getting full content from {self._selected_engine_name}: {str(e)}")
                # Fall back to returning relevant items without full content
                return relevant_items
        else:
            logger.warning("No engine was selected during preview phase, returning relevant items as-is")
            return relevant_items
    
    def _get_engine_instance(self, engine_name: str) -> Optional[BaseSearchEngine]:
        """Get or create an instance of the specified search engine"""
        # Return cached instance if available
        if engine_name in self.engine_cache:
            return self.engine_cache[engine_name]
        
        # Create a new instance
        engine = None
        try:
            # Only pass parameters that all engines accept
            common_params = {
                "llm": self.llm,
                "max_results": self.max_results
            }
            
            # Add max_filtered_results if specified
            if self.max_filtered_results is not None:
                common_params["max_filtered_results"] = self.max_filtered_results
                
            engine = create_search_engine(
                engine_name, 
                **common_params
            )
        except Exception as e:
            logger.error(f"Error creating engine instance for {engine_name}: {str(e)}")
            return None
        
        if engine:
            # Cache the instance
            self.engine_cache[engine_name] = engine
        
        return engine
    
    def invoke(self, query: str) -> List[Dict[str, Any]]:
        """Compatibility method for LangChain tools"""
        return self.run(query)


from typing import Dict, List, Any, Optional
from langchain_core.language_models import BaseLLM

from local_deep_research.web_search_engines.search_engine_base import BaseSearchEngine
from local_deep_research import config
import arxiv
import logging
logger = logging.getLogger(__name__)

class ArXivSearchEngine(BaseSearchEngine):
    """arXiv search engine implementation with two-phase approach"""
    
    def __init__(self, 
                max_results: int = 10, 
                sort_by: str = "relevance",
                sort_order: str = "descending",
                include_full_text: bool = False,
                download_dir: Optional[str] = None,
                max_full_text: int = 1,
                llm: Optional[BaseLLM] = None,
                max_filtered_results: Optional[int] = None):  # Added this parameter
        """
        Initialize the arXiv search engine.
        
        Args:
            max_results: Maximum number of search results
            sort_by: Sorting criteria ('relevance', 'lastUpdatedDate', or 'submittedDate')
            sort_order: Sort order ('ascending' or 'descending')
            include_full_text: Whether to include full paper content in results (downloads PDF)
            download_dir: Directory to download PDFs to (if include_full_text is True)
            max_full_text: Maximum number of PDFs to download and process (default: 1)
            llm: Language model for relevance filtering
            max_filtered_results: Maximum number of results to keep after filtering
        """
        # Initialize the BaseSearchEngine with LLM, max_filtered_results, and max_results
        super().__init__(llm=llm, max_filtered_results=max_filtered_results, max_results=max_results)
        self.max_results=max(self.max_results,25)
        self.sort_by = sort_by
        self.sort_order = sort_order
        self.include_full_text = include_full_text
        self.download_dir = download_dir
        self.max_full_text = max_full_text
        
        # Map sort parameters to arxiv package parameters
        self.sort_criteria = {
            'relevance': arxiv.SortCriterion.Relevance,
            'lastUpdatedDate': arxiv.SortCriterion.LastUpdatedDate,
            'submittedDate': arxiv.SortCriterion.SubmittedDate
        }
        
        self.sort_directions = {
            'ascending': arxiv.SortOrder.Ascending,
            'descending': arxiv.SortOrder.Descending
        }

    def _get_search_results(self, query: str) -> List[Any]:
        """
        Helper method to get search results from arXiv API.
        
        Args:
            query: The search query
            
        Returns:
            List of arXiv paper objects
        """
        # Configure the search client
        sort_criteria = self.sort_criteria.get(self.sort_by, arxiv.SortCriterion.Relevance)
        sort_order = self.sort_directions.get(self.sort_order, arxiv.SortOrder.Descending)
        
        
        
        # Create the search client
        client = arxiv.Client(page_size=self.max_results)
        
        # Create the search query
        search = arxiv.Search(
            query=query,
            max_results=self.max_results,
            sort_by=sort_criteria,
            sort_order=sort_order
        )
        
        # Get the search results
        papers = list(client.results(search))
        
        return papers
    
    def _get_previews(self, query: str) -> List[Dict[str, Any]]:
        """
        Get preview information for arXiv papers.
        
        Args:
            query: The search query
            
        Returns:
            List of preview dictionaries
        """
        logger.info("Getting paper previews from arXiv")
        
        try:
            # Get search results from arXiv
            papers = self._get_search_results(query)
            
            # Store the paper objects for later use
            self._papers = {paper.entry_id: paper for paper in papers}
            
            # Format results as previews with basic information
            previews = []
            for paper in papers:
                preview = {
                    "id": paper.entry_id,  # Use entry_id as ID
                    "title": paper.title,
                    "link": paper.entry_id,  # arXiv URL
                    "snippet": paper.summary[:250] + "..." if len(paper.summary) > 250 else paper.summary,
                    "authors": [author.name for author in paper.authors[:3]],  # First 3 authors
                    "published": paper.published.strftime("%Y-%m-%d") if paper.published else None
                }
                
                previews.append(preview)
            
            return previews
            
        except Exception as e:
            logger.error(f"Error getting arXiv previews: {e}")
            return []
    
    def _get_full_content(self, relevant_items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Get full content for the relevant arXiv papers.
        Downloads PDFs and extracts text when include_full_text is True.
        Limits the number of PDFs processed to max_full_text.
        
        Args:
            relevant_items: List of relevant preview dictionaries
            
        Returns:
            List of result dictionaries with full content
        """
        # Check if we should get full content
        if hasattr(config, 'SEARCH_SNIPPETS_ONLY') and config.SEARCH_SNIPPETS_ONLY:
            logger.info("Snippet-only mode, skipping full content retrieval")
            return relevant_items
            
        logger.info("Getting full content for relevant arXiv papers")
        
        results = []
        pdf_count = 0  # Track number of PDFs processed
        
        for item in relevant_items:
            # Start with the preview data
            result = item.copy()
            
            # Get the paper ID
            paper_id = item.get("id")
            
            # Try to get the full paper from our cache
            paper = None
            if hasattr(self, '_papers') and paper_id in self._papers:
                paper = self._papers[paper_id]
            
            if paper:
                # Add complete paper information
                result.update({
                    "pdf_url": paper.pdf_url,
                    "authors": [author.name for author in paper.authors],  # All authors
                    "published": paper.published.strftime("%Y-%m-%d") if paper.published else None,
                    "updated": paper.updated.strftime("%Y-%m-%d") if paper.updated else None,
                    "categories": paper.categories,
                    "summary": paper.summary,  # Full summary
                    "comment": paper.comment,
                    "journal_ref": paper.journal_ref,
                    "doi": paper.doi
                })
                
                # Default to using summary as content
                result["content"] = paper.summary
                result["full_content"] = paper.summary
                
                # Download PDF and extract text if requested and within limit
                if self.include_full_text and self.download_dir and pdf_count < self.max_full_text:
                    try:
                        # Download the paper
                        pdf_count += 1  # Increment counter before attempting download
                        paper_path = paper.download_pdf(dirpath=self.download_dir)
                        result["pdf_path"] = str(paper_path)
                        
                        # Extract text from PDF
                        try:
                            # Try PyPDF2 first
                            try:
                                import PyPDF2
                                with open(paper_path, 'rb') as pdf_file:
                                    pdf_reader = PyPDF2.PdfReader(pdf_file)
                                    pdf_text = ""
                                    for page in pdf_reader.pages:
                                        pdf_text += page.extract_text() + "\n\n"
                                    
                                    if pdf_text.strip():  # Only use if we got meaningful text
                                        result["content"] = pdf_text
                                        result["full_content"] = pdf_text
                                        logger.info(f"Successfully extracted text from PDF using PyPDF2")
                            except (ImportError, Exception) as e1:
                                # Fall back to pdfplumber
                                try:
                                    import pdfplumber
                                    with pdfplumber.open(paper_path) as pdf:
                                        pdf_text = ""
                                        for page in pdf.pages:
                                            pdf_text += page.extract_text() + "\n\n"
                                        
                                        if pdf_text.strip():  # Only use if we got meaningful text
                                            result["content"] = pdf_text
                                            result["full_content"] = pdf_text
                                            logger.info(f"Successfully extracted text from PDF using pdfplumber")
                                except (ImportError, Exception) as e2:
                                    logger.error(f"PDF text extraction failed: {str(e1)}, then {str(e2)}")
                                    logger.error(f"Using paper summary as content instead")
                        except Exception as e:
                            logger.error(f"Error extracting text from PDF: {e}")
                            logger.error(f"Using paper summary as content instead")
                    except Exception as e:
                        logger.error(f"Error downloading paper {paper.title}: {e}")
                        result["pdf_path"] = None
                        pdf_count -= 1  # Decrement counter if download fails
                elif self.include_full_text and self.download_dir and pdf_count >= self.max_full_text:
                    # Reached PDF limit
                    logger.info(f"Maximum number of PDFs ({self.max_full_text}) reached. Skipping remaining PDFs.")
                    result["content"] = paper.summary
                    result["full_content"] = paper.summary
            
            results.append(result)
        
        return results
    
    def run(self, query: str) -> List[Dict[str, Any]]:
        """
        Execute a search using arXiv with the two-phase approach.
        
        Args:
            query: The search query
            
        Returns:
            List of search results
        """
        logger.info("---Execute a search using arXiv---")
        
        # Use the implementation from the parent class which handles all phases
        results = super().run(query)
        
        # Clean up
        if hasattr(self, '_papers'):
            del self._papers
            
        return results
    
    def get_paper_details(self, arxiv_id: str) -> Dict[str, Any]:
        """
        Get detailed information about a specific arXiv paper.
        
        Args:
            arxiv_id: arXiv ID of the paper (e.g., '2101.12345')
            
        Returns:
            Dictionary with paper information
        """
        try:
            # Create the search client
            client = arxiv.Client()
            
            # Search for the specific paper
            search = arxiv.Search(id_list=[arxiv_id], max_results=1)
            
            # Get the paper
            papers = list(client.results(search))
            if not papers:
                return {}
            
            paper = papers[0]
            
            # Format result based on config
            result = {
                "title": paper.title,
                "link": paper.entry_id,
                "snippet": paper.summary[:250] + "..." if len(paper.summary) > 250 else paper.summary,
                "authors": [author.name for author in paper.authors[:3]]  # First 3 authors
            }
            
            # Add full content if not in snippet-only mode
            if not hasattr(config, 'SEARCH_SNIPPETS_ONLY') or not config.SEARCH_SNIPPETS_ONLY:
                result.update({
                    "pdf_url": paper.pdf_url,
                    "authors": [author.name for author in paper.authors],  # All authors
                    "published": paper.published.strftime("%Y-%m-%d") if paper.published else None,
                    "updated": paper.updated.strftime("%Y-%m-%d") if paper.updated else None,
                    "categories": paper.categories,
                    "summary": paper.summary,  # Full summary
                    "comment": paper.comment,
                    "journal_ref": paper.journal_ref,
                    "doi": paper.doi,
                    "content": paper.summary,  # Use summary as content
                    "full_content": paper.summary  # For consistency
                })
                
                # Download PDF if requested
                if self.include_full_text and self.download_dir:
                    try:
                        # Download the paper
                        paper_path = paper.download_pdf(dirpath=self.download_dir)
                        result["pdf_path"] = str(paper_path)
                    except Exception as e:
                        logger.error(f"Error downloading paper: {e}")
            
            return result
            
        except Exception as e:
            logger.error(f"Error getting paper details: {e}")
            return {}
    
    def search_by_author(self, author_name: str, max_results: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Search for papers by a specific author.
        
        Args:
            author_name: Name of the author
            max_results: Maximum number of results (defaults to self.max_results)
            
        Returns:
            List of papers by the author
        """
        original_max_results = self.max_results
        
        try:
            if max_results:
                self.max_results = max_results
                
            query = f"au:\"{author_name}\""
            return self.run(query)
            
        finally:
            # Restore original value
            self.max_results = original_max_results
    
    def search_by_category(self, category: str, max_results: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Search for papers in a specific arXiv category.
        
        Args:
            category: arXiv category (e.g., 'cs.AI', 'physics.optics')
            max_results: Maximum number of results (defaults to self.max_results)
            
        Returns:
            List of papers in the category
        """
        original_max_results = self.max_results
        
        try:
            if max_results:
                self.max_results = max_results
                
            query = f"cat:{category}"
            return self.run(query)
            
        finally:
            # Restore original value
            self.max_results = original_max_results


"""
Search engine that searches across all local collections
"""

import logging
from typing import Dict, List, Any, Optional

import toml
from langchain_core.language_models import BaseLLM

from local_deep_research.web_search_engines.search_engine_base import BaseSearchEngine
from local_deep_research.web_search_engines.search_engine_factory import create_search_engine
from local_deep_research.config import LOCAL_COLLECTIONS_FILE

# Setup logging
logger = logging.getLogger(__name__)

class LocalAllSearchEngine(BaseSearchEngine):
    """
    Search engine that searches across all local document collections.
    Acts as a meta search engine specifically for local collections.
    """

    def __init__(
        self,
        llm: Optional[BaseLLM] = None,
        max_results: int = 10,
        max_filtered_results: Optional[int] = None,
        **kwargs
    ):
        """
        Initialize the local all-collections search engine.
        
        Args:
            llm: Language model for relevance filtering
            max_results: Maximum number of search results
            max_filtered_results: Maximum results after filtering
            **kwargs: Additional parameters passed to LocalSearchEngine instances
        """
        # Initialize the BaseSearchEngine with LLM, max_filtered_results, and max_results
        super().__init__(llm=llm, max_filtered_results=max_filtered_results, max_results=max_results)
                
        # Find all local collection search engines
        self.local_engines = {}
        try:
            local_collections = toml.load(LOCAL_COLLECTIONS_FILE)

            for collection_id, collection in local_collections.items():
                if not collection.get("enabled", True):
                    continue
                    
                # Create a search engine for this collection
                try:
                    engine = create_search_engine(
                        collection_id,
                        llm=llm,
                        max_filtered_results=max_filtered_results
                    )
                    
                    if engine:
                        self.local_engines[collection_id] = {
                            "engine": engine,
                            "name": collection.get("name", collection_id),
                            "description": collection.get("description", "")
                        }
                except Exception as e:
                    logger.error(f"Error creating search engine for collection '{collection_id}': {e}")
        except ImportError:
            logger.warning("No local collections configuration found")
    
    def _get_previews(self, query: str) -> List[Dict[str, Any]]:
        """
        Get preview information for documents from all local collections.
        
        Args:
            query: The search query
            
        Returns:
            List of preview dictionaries
        """
        logger.info(f"Searching across all local collections for query: {query}")
        
        all_previews = []
        
        # Get previews from each local search engine
        for collection_id, engine_info in self.local_engines.items():
            engine = engine_info["engine"]
            try:
                # Get previews from this engine
                previews = engine._get_previews(query)
                
                # Add collection info to each preview
                for preview in previews:
                    preview["collection_id"] = collection_id
                    preview["collection_name"] = engine_info["name"]
                    preview["collection_description"] = engine_info["description"]
                
                all_previews.extend(previews)
            except Exception as e:
                logger.error(f"Error searching collection '{collection_id}': {e}")
        
        if not all_previews:
            logger.info(f"No local documents found for query: {query}")
            return []
        
        # Sort by similarity score if available
        all_previews.sort(
            key=lambda x: float(x.get("similarity", 0)), 
            reverse=True
        )
        
        # Limit to max_results
        return all_previews[:self.max_results]
    
    def _get_full_content(self, relevant_items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Get full content for the relevant documents.
        Delegates to the appropriate collection's search engine.
        
        Args:
            relevant_items: List of relevant preview dictionaries
            
        Returns:
            List of result dictionaries with full content
        """
        # Group items by collection
        items_by_collection = {}
        for item in relevant_items:
            collection_id = item.get("collection_id")
            if collection_id and collection_id in self.local_engines:
                if collection_id not in items_by_collection:
                    items_by_collection[collection_id] = []
                items_by_collection[collection_id].append(item)
        
        # Process each collection's items with its own engine
        all_results = []
        for collection_id, items in items_by_collection.items():
            engine = self.local_engines[collection_id]["engine"]
            try:
                results = engine._get_full_content(items)
                all_results.extend(results)
            except Exception as e:
                logger.error(f"Error getting full content from collection '{collection_id}': {e}")
                # Fall back to returning the items without full content
                all_results.extend(items)
        
        # Add any items that weren't processed
        processed_ids = set(item["id"] for item in all_results)
        for item in relevant_items:
            if item["id"] not in processed_ids:
                all_results.append(item)
        
        return all_results



import requests
import logging
import os
from typing import Dict, List, Any, Optional
from langchain_core.language_models import BaseLLM
import time
import json

from web_search_engines.search_engine_base import BaseSearchEngine
from web_search_engines.engines.full_search import FullSearchResults
import config

# Setup logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class SearXNGSearchEngine(BaseSearchEngine):
    """
    SearXNG search engine implementation that requires an instance URL provided via
    environment variable or configuration. Designed for ethical usage with proper
    rate limiting and single-instance approach.
    """
    
    def __init__(self, 
                max_results: int = 15,
                instance_url: Optional[str] = None,  # Can be None if using env var
                categories: Optional[List[str]] = None,
                engines: Optional[List[str]] = None,
                language: str = "en",
                safe_search: int = 1,
                time_range: Optional[str] = None,
                delay_between_requests: float = 2.0,
                llm: Optional[BaseLLM] = None,
                max_filtered_results: Optional[int] = None,
                include_full_content: bool = True,
                api_key: Optional[str] = None):  # API key is actually the instance URL
        """
        Initialize the SearXNG search engine with ethical usage patterns.
        
        Args:
            max_results: Maximum number of search results
            instance_url: URL of your SearXNG instance (preferably self-hosted)
            categories: List of SearXNG categories to search in (general, images, videos, news, etc.)
            engines: List of engines to use (google, bing, duckduckgo, etc.)
            language: Language code for search results
            safe_search: Safe search level (0=off, 1=moderate, 2=strict)
            time_range: Time range for results (day, week, month, year)
            delay_between_requests: Seconds to wait between requests
            llm: Language model for relevance filtering
            max_filtered_results: Maximum number of results to keep after filtering
            include_full_content: Whether to include full webpage content in results
            api_key: Alternative way to provide instance URL (takes precedence over instance_url)
        """

        # Initialize the BaseSearchEngine with LLM, max_filtered_results, and max_results
        super().__init__(llm=llm, max_filtered_results=max_filtered_results, max_results=max_results)
        
        # Get instance URL from various sources in priority order:
        # 1. api_key parameter (which is actually the instance URL)
        # 2. SEARXNG_INSTANCE environment variable
        # 3. instance_url parameter
        # 4. Default to None, which will disable the engine
        self.instance_url = api_key or os.getenv("SEARXNG_INSTANCE") or instance_url
        
        # Add debug logging for instance URL
        logger.info(f"SearXNG init - Instance URL sources: api_key={api_key}, env={os.getenv('SEARXNG_INSTANCE')}, param={instance_url}")
        
        # Validate and normalize the instance URL if provided
        if self.instance_url:
            self.instance_url = self.instance_url.rstrip('/')
            self.is_available = True
            logger.info(f"SearXNG initialized with instance URL: {self.instance_url}")
        else:
            self.is_available = False
            logger.error("No SearXNG instance URL provided. The engine is disabled. "
                       "Set SEARXNG_INSTANCE environment variable or provide instance_url parameter.")
        
        # Add debug logging for all parameters
        logger.info(f"SearXNG init params: max_results={max_results}, language={language}, "
                    f"max_filtered_results={max_filtered_results}, is_available={self.is_available}")
        
        self.max_results = max_results
        self.categories = categories or ["general"]
        self.engines = engines
        self.language = language
        self.safe_search = safe_search
        self.time_range = time_range
        
        self.delay_between_requests = float(os.getenv("SEARXNG_DELAY", delay_between_requests))
        
        self.include_full_content = include_full_content
        
        if self.is_available:
            self.search_url = f"{self.instance_url}/search"
            logger.info(f"SearXNG engine initialized with instance: {self.instance_url}")
            logger.info(f"Rate limiting set to {self.delay_between_requests} seconds between requests")
        
            self.full_search = FullSearchResults(
                llm=llm,
                web_search=self,
                language=language,
                max_results=max_results,
                region="wt-wt",
                time="y",
                safesearch="Moderate" if safe_search == 1 else "Off" if safe_search == 0 else "Strict"
            )
        
        self.last_request_time = 0
    
    def _respect_rate_limit(self):
        """Apply self-imposed rate limiting between requests"""
        current_time = time.time()
        time_since_last_request = current_time - self.last_request_time
        

        if time_since_last_request < self.delay_between_requests:
            wait_time = self.delay_between_requests - time_since_last_request
            logger.info(f"Rate limiting: waiting {wait_time:.2f} seconds")
            time.sleep(wait_time)
        
        self.last_request_time = time.time()
    
    def _get_search_results(self, query: str) -> List[Dict[str, Any]]:
        """
        Get search results from SearXNG with ethical rate limiting.
        
        Args:
            query: The search query
            
        Returns:
            List of search results from SearXNG
        """
        if not self.is_available:
            logger.error("SearXNG engine is disabled (no instance URL provided) - cannot run search")
            return []
            
        logger.info(f"SearXNG running search for query: {query}")
        
        try:
            self._respect_rate_limit()
            
            initial_headers = {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                "Accept-Language": "en-US,en;q=0.9"
            }
            
            try:
                initial_response = requests.get(self.instance_url, headers=initial_headers, timeout=10)
                cookies = initial_response.cookies
            except Exception as e:
                logger.warning(f"Failed to get initial cookies: {e}")
                cookies = None
            
            params = {
                "q": query,
                "categories": ",".join(self.categories),
                "language": self.language,
                "format": "html",  # Use HTML format instead of JSON
                "pageno": 1,
                "safesearch": self.safe_search,
                "count": self.max_results
            }
            
            if self.engines:
                params["engines"] = ",".join(self.engines)
                
            if self.time_range:
                params["time_range"] = self.time_range
            
            # Browser-like headers
            headers = {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                "Accept-Language": "en-US,en;q=0.9",
                "Referer": self.instance_url + "/",
                "Connection": "keep-alive",
                "Upgrade-Insecure-Requests": "1"
            }
            
            logger.info(f"Sending request to SearXNG instance at {self.instance_url}")
            response = requests.get(
                self.search_url,
                params=params,
                headers=headers,
                cookies=cookies,
                timeout=15
            )
            
            if response.status_code == 200:
                try:
                    from bs4 import BeautifulSoup
                    
                    soup = BeautifulSoup(response.text, 'html.parser')
                    results = []
                    
                    result_elements = soup.select('.result-item')
                    
                    if not result_elements:
                        result_elements = soup.select('.result')
                    
                    if not result_elements:
                        result_elements = soup.select('article') 
                        
                    if not result_elements:
                        logger.debug(f"Classes found in HTML: {[c['class'] for c in soup.select('[class]') if 'class' in c.attrs][:10]}")
                        result_elements = soup.select('div[id^="result"]')
                    
                    logger.info(f"Found {len(result_elements)} search result elements")
                        
                    for idx, result_element in enumerate(result_elements):
                        if idx >= self.max_results:
                            break
                            
                        title_element = (
                            result_element.select_one('.result-title') or 
                            result_element.select_one('.title') or 
                            result_element.select_one('h3') or
                            result_element.select_one('a[href]')
                        )
                        
                        url_element = (
                            result_element.select_one('.result-url') or 
                            result_element.select_one('.url') or
                            result_element.select_one('a[href]')
                        )
                        
                        content_element = (
                            result_element.select_one('.result-content') or
                            result_element.select_one('.content') or
                            result_element.select_one('.snippet') or 
                            result_element.select_one('p')
                        )
                        
                        title = title_element.get_text(strip=True) if title_element else ""
                        
                        url = ""
                        if url_element and url_element.has_attr('href'):
                            url = url_element['href']
                        elif url_element:
                            url = url_element.get_text(strip=True)
                            
                        content = content_element.get_text(strip=True) if content_element else ""
                        
                        if not url and title_element and title_element.has_attr('href'):
                            url = title_element['href']
                            
                        logger.debug(f"Extracted result {idx}: title={title[:30]}..., url={url[:30]}..., content={content[:30]}...")
                        
                        # Add to results if we have at least a title or URL
                        if title or url:
                            results.append({
                                "title": title,
                                "url": url,
                                "content": content,
                                "engine": "searxng",
                                "category": "general"
                            })
                    
                    logger.info(f"SearXNG returned {len(results)} results from HTML parsing")
                    return results
                    
                except ImportError:
                    logger.error("BeautifulSoup not available for HTML parsing")
                    return []
                except Exception as e:
                    logger.error(f"Error parsing HTML results: {str(e)}")
                    return []
            else:
                logger.error(f"SearXNG returned status code {response.status_code}")
                return []
                
        except Exception as e:
            logger.error(f"Error getting SearXNG results: {e}")
            return []
    
    def _get_previews(self, query: str) -> List[Dict[str, Any]]:
        """
        Get preview information for SearXNG search results.
        
        Args:
            query: The search query
            
        Returns:
            List of preview dictionaries
        """
        if not self.is_available:
            logger.warning("SearXNG engine is disabled (no instance URL provided)")
            return []
            
        logger.info(f"Getting SearXNG previews for query: {query}")
        
        results = self._get_search_results(query)
        
        if not results:
            logger.warning(f"No SearXNG results found for query: {query}")
            return []
        
        previews = []
        for i, result in enumerate(results):
            title = result.get("title", "")
            url = result.get("url", "")
            content = result.get("content", "")
            
            preview = {
                "id": url or f"searxng-result-{i}",
                "title": title,
                "link": url,
                "snippet": content,
                "engine": result.get("engine", ""),
                "category": result.get("category", "")
            }
            
            previews.append(preview)
        
        return previews
    
    def _get_full_content(self, relevant_items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Get full content for the relevant search results.
        
        Args:
            relevant_items: List of relevant preview dictionaries
            
        Returns:
            List of result dictionaries with full content
        """
        if not self.is_available:
            return relevant_items
            
        if hasattr(config, 'SEARCH_SNIPPETS_ONLY') and config.SEARCH_SNIPPETS_ONLY:
            logger.info("Snippet-only mode, skipping full content retrieval")
            return relevant_items
        
        logger.info("Retrieving full webpage content")
        
        try:
            results_with_content = self.full_search._get_full_content(relevant_items)
            return results_with_content
            
        except Exception as e:
            logger.error(f"Error retrieving full content: {e}")
            return relevant_items
    
    def invoke(self, query: str) -> List[Dict[str, Any]]:
        """Compatibility method for LangChain tools"""
        return self.run(query)
    
    def results(self, query: str, max_results: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Get search results in a format compatible with other search engines.
        
        Args:
            query: The search query
            max_results: Optional override for maximum results
            
        Returns:
            List of search result dictionaries
        """
        if not self.is_available:
            return []
            
        original_max_results = self.max_results
        
        try:
            if max_results is not None:
                self.max_results = max_results
                
            results = self._get_search_results(query)
            
            formatted_results = []
            for result in results:
                formatted_results.append({
                    "title": result.get("title", ""),
                    "link": result.get("url", ""),
                    "snippet": result.get("content", "")
                })
                
            return formatted_results
            
        finally:
            self.max_results = original_max_results
    
    @staticmethod
    def get_self_hosting_instructions() -> str:
        """
        Get instructions for self-hosting a SearXNG instance.
        
        Returns:
            String with installation instructions
        """
        return """
# SearXNG Self-Hosting Instructions

The most ethical way to use SearXNG is to host your own instance. Here's how:

## Using Docker (easiest method)

1. Install Docker if you don't have it already
2. Run these commands:

```bash
# Pull the SearXNG Docker image
docker pull searxng/searxng

# Run SearXNG (will be available at http://localhost:8080)
docker run -d -p 8080:8080 --name searxng searxng/searxng
```

## Using Docker Compose (recommended for production)

1. Create a file named `docker-compose.yml` with the following content:

```yaml
version: '3'
services:
  searxng:
    container_name: searxng
    image: searxng/searxng
    ports:
      - "8080:8080"
    volumes:
      - ./searxng:/etc/searxng
    environment:
      - SEARXNG_BASE_URL=http://localhost:8080/
    restart: unless-stopped
```

2. Run with Docker Compose:

```bash
docker-compose up -d
```

For more detailed instructions and configuration options, visit:
https://searxng.github.io/searxng/admin/installation.html
"""

    def run(self, query: str) -> List[Dict[str, Any]]:
        """
        Override BaseSearchEngine run method to add SearXNG-specific error handling.
        """
        if not self.is_available:
            logger.error("SearXNG run method called but engine is not available (missing instance URL)")
            return []
        
        logger.info(f"SearXNG run method called with query: {query}")
        
        try:
            # Call the parent class's run method
            return super().run(query)
        except Exception as e:
            logger.error(f"Error in SearXNG run method: {str(e)}")
            # Return empty results on error
            return []


import os
import json
import time
import sqlite3
import threading
from datetime import datetime
from flask import Flask, render_template, request, jsonify, send_from_directory, Response, make_response, current_app, Blueprint, redirect, url_for, flash
from flask_socketio import SocketIO, emit
from local_deep_research.search_system import AdvancedSearchSystem
from local_deep_research.report_generator import IntegratedReportGenerator
# Move this import up to ensure it's available globally
from dateutil import parser
import traceback
import pkg_resources
# Import the new configuration manager
from local_deep_research.config import get_config_dir 
import logging
logger = logging.getLogger(__name__)

CONFIG_DIR = get_config_dir() / "config"
MAIN_CONFIG_FILE = CONFIG_DIR / "settings.toml"
LLM_CONFIG_FILE = CONFIG_DIR / "llm_config.py"
LOCAL_COLLECTIONS_FILE = CONFIG_DIR / "local_collections.toml"
import toml

# Set flag for tracking OpenAI availability - we'll check it only when needed
OPENAI_AVAILABLE = False

# Initialize Flask app
try:
    import os
    import logging
    from local_deep_research.utilties.setup_utils import setup_user_directories
    
    # Configure logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)
    
    # Explicitly run setup
    logger.info("Initializing configuration...")
    setup_user_directories()
    
    # Get directories based on package installation
    PACKAGE_DIR = pkg_resources.resource_filename('local_deep_research', 'web')
    STATIC_DIR = os.path.join(PACKAGE_DIR, 'static')
    TEMPLATE_DIR = os.path.join(PACKAGE_DIR, 'templates')
    
    # Setup logging
    logging.basicConfig(level=logging.INFO)
    
    # Create directories and default configs if needed
    setup_user_directories()

    # Initialize Flask app with package directories
    app = Flask(__name__, 
                static_folder=STATIC_DIR,
                template_folder=TEMPLATE_DIR)
    print(f"Using package static path: {STATIC_DIR}")
    print(f"Using package template path: {TEMPLATE_DIR}")
except Exception as e:
    # Fallback for development
    print(f"Package directories not found, using fallback paths: {str(e)}")
    app = Flask(__name__, 
                static_folder=os.path.abspath('static'),
                template_folder=os.path.abspath('templates'))
app.config['SECRET_KEY'] = 'deep-research-secret-key'

# Create a Blueprint for the research application
research_bp = Blueprint('research', __name__, url_prefix='/research')

# Add improved Socket.IO configuration with better error handling
socketio = SocketIO(
    app, 
    cors_allowed_origins="*",
    async_mode='threading',
    path='/research/socket.io',
    logger=True,
    engineio_logger=True,
    ping_timeout=20,
    ping_interval=5
)

# Active research processes and socket subscriptions
active_research = {}
socket_subscriptions = {}

# Add termination flags dictionary
termination_flags = {}

# Database setup
DB_PATH = 'research_history.db'

# Output directory for research results
OUTPUT_DIR = 'research_outputs'

# Add Content Security Policy headers to allow Socket.IO to function
@app.after_request
def add_security_headers(response):
    # Define a permissive CSP for development that allows Socket.IO to function
    csp = (
        "default-src 'self'; "
        "connect-src 'self' ws: wss: http: https:; " 
        "script-src 'self' 'unsafe-inline' 'unsafe-eval' cdnjs.cloudflare.com cdn.jsdelivr.net unpkg.com; "
        "style-src 'self' 'unsafe-inline' cdnjs.cloudflare.com; "
        "font-src 'self' cdnjs.cloudflare.com; "
        "img-src 'self' data:; "
        "worker-src blob:; "
        "frame-src 'self';"
    )
    
    response.headers['Content-Security-Policy'] = csp
    response.headers['X-Content-Security-Policy'] = csp
    
    # Add CORS headers for API requests
    if request.path.startswith('/api/'):
        response.headers['Access-Control-Allow-Origin'] = '*'
        response.headers['Access-Control-Allow-Methods'] = 'GET, POST, DELETE, OPTIONS'
        response.headers['Access-Control-Allow-Headers'] = 'Content-Type'
    
    return response

# Add a middleware layer to handle abrupt disconnections
@app.before_request
def handle_websocket_requests():
    if request.path.startswith('/research/socket.io'):
        try:
            if not request.environ.get('werkzeug.socket'):
                return
        except Exception as e:
            print(f"WebSocket preprocessing error: {e}")
            # Return empty response to prevent further processing
            return '', 200

def init_db():
    conn = sqlite3.connect(DB_PATH)
    cursor = conn.cursor()
    
    # Create the table if it doesn't exist
    cursor.execute('''
    CREATE TABLE IF NOT EXISTS research_history (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        query TEXT NOT NULL,
        mode TEXT NOT NULL,
        status TEXT NOT NULL,
        created_at TEXT NOT NULL,
        completed_at TEXT,
        duration_seconds INTEGER,
        report_path TEXT,
        metadata TEXT,
        progress_log TEXT,
        progress INTEGER
    )
    ''')
    
    # Create a dedicated table for research logs
    cursor.execute('''
    CREATE TABLE IF NOT EXISTS research_logs (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        research_id INTEGER NOT NULL,
        timestamp TEXT NOT NULL,
        message TEXT NOT NULL,
        log_type TEXT NOT NULL,
        progress INTEGER,
        metadata TEXT,
        FOREIGN KEY (research_id) REFERENCES research_history (id) ON DELETE CASCADE
    )
    ''')
    
    # Check if the duration_seconds column exists, add it if missing
    cursor.execute('PRAGMA table_info(research_history)')
    columns = [column[1] for column in cursor.fetchall()]
    
    if 'duration_seconds' not in columns:
        print("Adding missing 'duration_seconds' column to research_history table")
        cursor.execute('ALTER TABLE research_history ADD COLUMN duration_seconds INTEGER')
    
    # Check if the progress column exists, add it if missing
    if 'progress' not in columns:
        print("Adding missing 'progress' column to research_history table")
        cursor.execute('ALTER TABLE research_history ADD COLUMN progress INTEGER')
    
    # Enable foreign key support
    cursor.execute('PRAGMA foreign_keys = ON')
    
    conn.commit()
    conn.close()
    
# Helper function to calculate duration between created_at and completed_at timestamps
def calculate_duration(created_at_str):
    """
    Calculate duration in seconds between created_at timestamp and now.
    Handles various timestamp formats and returns None if calculation fails.
    """
    if not created_at_str:
        return None
        
    now = datetime.utcnow()
    duration_seconds = None
    
    try:
        # Proper parsing of ISO format
        if 'T' in created_at_str:  # ISO format with T separator
            start_time = datetime.fromisoformat(created_at_str)
        else:  # Older format without T
            # Try different formats
            try:
                start_time = datetime.strptime(created_at_str, '%Y-%m-%d %H:%M:%S.%f')
            except ValueError:
                try:
                    start_time = datetime.strptime(created_at_str, '%Y-%m-%d %H:%M:%S')
                except ValueError:
                    # Last resort fallback
                    start_time = datetime.fromisoformat(created_at_str.replace(' ', 'T'))
        
        # Ensure we're comparing UTC times
        duration_seconds = int((now - start_time).total_seconds())
    except Exception as e:
        print(f"Error calculating duration: {str(e)}")
        # Fallback method if parsing fails
        try:
            start_time_fallback = parser.parse(created_at_str)
            duration_seconds = int((now - start_time_fallback).total_seconds())
        except:
            print(f"Fallback duration calculation also failed for timestamp: {created_at_str}")
    
    return duration_seconds

# Add these helper functions after the calculate_duration function


def add_log_to_db(research_id, message, log_type='info', progress=None, metadata=None):
    """
    Store a log entry in the database
    
    Args:
        research_id: ID of the research
        message: Log message text
        log_type: Type of log (info, error, milestone)
        progress: Progress percentage (0-100)
        metadata: Additional metadata as dictionary (will be stored as JSON)
    """
    try:
        timestamp = datetime.utcnow().isoformat()
        metadata_json = json.dumps(metadata) if metadata else None
        
        conn = sqlite3.connect(DB_PATH)
        cursor = conn.cursor()
        cursor.execute(
            'INSERT INTO research_logs (research_id, timestamp, message, log_type, progress, metadata) '
            'VALUES (?, ?, ?, ?, ?, ?)',
            (research_id, timestamp, message, log_type, progress, metadata_json)
        )
        conn.commit()
        conn.close()
        return True
    except Exception as e:
        print(f"Error adding log to database: {str(e)}")
        print(traceback.format_exc())
        return False

def get_logs_for_research(research_id):
    """
    Retrieve all logs for a specific research ID
    
    Args:
        research_id: ID of the research
    
    Returns:
        List of log entries as dictionaries
    """
    try:
        conn = sqlite3.connect(DB_PATH)
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute(
            'SELECT * FROM research_logs WHERE research_id = ? ORDER BY timestamp ASC',
            (research_id,)
        )
        results = cursor.fetchall()
        conn.close()
        
        logs = []
        for result in results:
            log_entry = dict(result)
            # Parse metadata JSON if it exists
            if log_entry.get('metadata'):
                try:
                    log_entry['metadata'] = json.loads(log_entry['metadata'])
                except:
                    log_entry['metadata'] = {}
            else:
                log_entry['metadata'] = {}
            
            # Convert entry for frontend consumption
            formatted_entry = {
                'time': log_entry['timestamp'],
                'message': log_entry['message'],
                'progress': log_entry['progress'],
                'metadata': log_entry['metadata'],
                'type': log_entry['log_type']
            }
            logs.append(formatted_entry)
            
        return logs
    except Exception as e:
        print(f"Error retrieving logs from database: {str(e)}")
        print(traceback.format_exc())
        return []
    
# Initialize the database on startup
def initialize():
    init_db()

# Call initialize immediately when app is created
initialize()

# Route for index page - keep this at root level for easy access
@app.route('/')
def root_index():
    return redirect(url_for('research.index'))

# Update all routes with the research prefix
@research_bp.route('/')
def index():
    return render_template('index.html')

@research_bp.route('/static/<path:path>')
def serve_static(path):
    try:
        print(f"Serving static file: {path}")
        print(f"Static folder path: {app.static_folder}")
        return send_from_directory(app.static_folder, path)
    except Exception as e:
        print(f"Error serving static file {path}: {str(e)}")
        return f"Error serving file: {str(e)}", 404

@research_bp.route('/api/history', methods=['GET'])
def get_history():
    """Get the research history"""
    try:
        conn = sqlite3.connect(DB_PATH)
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        
        # Get all history records ordered by latest first
        cursor.execute('SELECT * FROM research_history ORDER BY created_at DESC')
        results = cursor.fetchall()
        conn.close()
        
        # Convert to list of dicts
        history = []
        for result in results:
            item = dict(result)
            
            # Ensure all keys exist with default values
            if 'id' not in item:
                item['id'] = None
            if 'query' not in item:
                item['query'] = 'Untitled Research'
            if 'mode' not in item:
                item['mode'] = 'quick'
            if 'status' not in item:
                item['status'] = 'unknown'
            if 'created_at' not in item:
                item['created_at'] = None
            if 'completed_at' not in item:
                item['completed_at'] = None
            if 'duration_seconds' not in item:
                item['duration_seconds'] = None
            if 'report_path' not in item:
                item['report_path'] = None
            if 'metadata' not in item:
                item['metadata'] = '{}'
            if 'progress_log' not in item:
                item['progress_log'] = '[]'
            
            # Ensure timestamps are in ISO format
            if item['created_at'] and 'T' not in item['created_at']:
                try:
                    # Convert to ISO format if it's not already
                    dt = parser.parse(item['created_at'])
                    item['created_at'] = dt.isoformat()
                except:
                    pass
                
            if item['completed_at'] and 'T' not in item['completed_at']:
                try:
                    # Convert to ISO format if it's not already
                    dt = parser.parse(item['completed_at'])
                    item['completed_at'] = dt.isoformat()
                except:
                    pass
                
            # Recalculate duration based on timestamps if it's null but both timestamps exist
            if item['duration_seconds'] is None and item['created_at'] and item['completed_at']:
                try:
                    start_time = parser.parse(item['created_at'])
                    end_time = parser.parse(item['completed_at'])
                    item['duration_seconds'] = int((end_time - start_time).total_seconds())
                except Exception as e:
                    print(f"Error recalculating duration: {str(e)}")
            
            history.append(item)
        
        # Add CORS headers
        response = make_response(jsonify(history))
        response.headers.add('Access-Control-Allow-Origin', '*')
        response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
        response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE,OPTIONS')
        return response
    except Exception as e:
        print(f"Error getting history: {str(e)}")
        print(traceback.format_exc())
        # Return empty array with CORS headers
        response = make_response(jsonify([]))
        response.headers.add('Access-Control-Allow-Origin', '*')
        response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
        response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE,OPTIONS')
        return response

@research_bp.route('/api/start_research', methods=['POST'])
def start_research():
    data = request.json
    query = data.get('query')
    mode = data.get('mode', 'quick')
    
    if not query:
        return jsonify({'status': 'error', 'message': 'Query is required'}), 400
        
    # Check if there's any active research that's actually still running
    if active_research:
        # Verify each active research is still valid
        stale_research_ids = []
        for research_id, research_data in list(active_research.items()):
            # Check database status
            conn = sqlite3.connect(DB_PATH)
            cursor = conn.cursor()
            cursor.execute('SELECT status FROM research_history WHERE id = ?', (research_id,))
            result = cursor.fetchone()
            conn.close()
            
            # If the research doesn't exist in DB or is not in_progress, it's stale
            if not result or result[0] != 'in_progress':
                stale_research_ids.append(research_id)
            # Also check if thread is still alive
            elif not research_data.get('thread') or not research_data.get('thread').is_alive():
                stale_research_ids.append(research_id)

        # Clean up any stale research processes
        for stale_id in stale_research_ids:
            print(f"Cleaning up stale research process: {stale_id}")
            if stale_id in active_research:
                del active_research[stale_id]
            if stale_id in termination_flags:
                del termination_flags[stale_id]

        # After cleanup, check if there's still active research
        if active_research:
            return jsonify({
                'status': 'error', 
                'message': 'Another research is already in progress. Please wait for it to complete.'
            }), 409
        
    # Create a record in the database with explicit UTC timestamp
    created_at = datetime.utcnow().isoformat()
    conn = sqlite3.connect(DB_PATH)
    cursor = conn.cursor()
    cursor.execute(
        'INSERT INTO research_history (query, mode, status, created_at, progress_log) VALUES (?, ?, ?, ?, ?)',
        (query, mode, 'in_progress', created_at, json.dumps([{"time": created_at, "message": "Research started", "progress": 0}]))
    )
    research_id = cursor.lastrowid
    conn.commit()
    conn.close()
    
    # Start research process in a background thread
    thread = threading.Thread(
        target=run_research_process,
        args=(research_id, query, mode)
    )
    thread.daemon = True
    thread.start()
    
    active_research[research_id] = {
        'thread': thread,
        'progress': 0,
        'status': 'in_progress',
        'log': [{"time": created_at, "message": "Research started", "progress": 0}]
    }
    
    return jsonify({
        'status': 'success',
        'research_id': research_id
    })

@research_bp.route('/api/research/<int:research_id>')
def get_research_status(research_id):
    conn = sqlite3.connect(DB_PATH)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()
    cursor.execute('SELECT * FROM research_history WHERE id = ?', (research_id,))
    result = dict(cursor.fetchone() or {})
    conn.close()
    
    if not result:
        return jsonify({'status': 'error', 'message': 'Research not found'}), 404
        
    # Add progress information
    if research_id in active_research:
        result['progress'] = active_research[research_id]['progress']
        result['log'] = active_research[research_id]['log']
    elif result.get('status') == 'completed':
        result['progress'] = 100
        try:
            result['log'] = json.loads(result.get('progress_log', '[]'))
        except:
            result['log'] = []
    else:
        result['progress'] = 0
        try:
            result['log'] = json.loads(result.get('progress_log', '[]'))
        except:
            result['log'] = []
        
    return jsonify(result)

@research_bp.route('/api/research/<int:research_id>/details')
def get_research_details(research_id):
    """Get detailed progress log for a specific research"""
    conn = sqlite3.connect(DB_PATH)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()
    cursor.execute('SELECT * FROM research_history WHERE id = ?', (research_id,))
    result = dict(cursor.fetchone() or {})
    conn.close()
    
    if not result:
        return jsonify({'status': 'error', 'message': 'Research not found'}), 404
    
    # Get logs from the dedicated log database
    logs = get_logs_for_research(research_id)
    
    # If this is an active research, merge with any in-memory logs
    if research_id in active_research:
        # Use the logs from memory temporarily until they're saved to the database
        memory_logs = active_research[research_id]['log']
        
        # Filter out logs that are already in the database by timestamp
        db_timestamps = {log['time'] for log in logs}
        unique_memory_logs = [log for log in memory_logs if log['time'] not in db_timestamps]
        
        # Add unique memory logs to our return list
        logs.extend(unique_memory_logs)
        
        # Sort logs by timestamp
        logs.sort(key=lambda x: x['time'])
    
    return jsonify({
        'status': 'success',
        'research_id': research_id,
        'query': result.get('query'),
        'mode': result.get('mode'),
        'status': result.get('status'),
        'progress': active_research.get(research_id, {}).get('progress', 100 if result.get('status') == 'completed' else 0),
        'created_at': result.get('created_at'),
        'completed_at': result.get('completed_at'),
        'log': logs
    })

@research_bp.route('/api/report/<int:research_id>')
def get_report(research_id):
    conn = sqlite3.connect(DB_PATH)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()
    cursor.execute('SELECT * FROM research_history WHERE id = ?', (research_id,))
    result = dict(cursor.fetchone() or {})
    conn.close()
    
    if not result or not result.get('report_path'):
        return jsonify({'status': 'error', 'message': 'Report not found'}), 404
        
    try:
        with open(result['report_path'], 'r', encoding='utf-8') as f:
            content = f.read()
        return jsonify({
            'status': 'success',
            'content': content,
            'metadata': json.loads(result.get('metadata', '{}'))
        })
    except Exception as e:
        return jsonify({'status': 'error', 'message': str(e)}), 500

@research_bp.route('/research/details/<int:research_id>')
def research_details_page(research_id):
    """Render the research details page"""
    return render_template('index.html')

@socketio.on('connect')
def handle_connect():
    print(f"Client connected: {request.sid}")

@socketio.on('disconnect')
def handle_disconnect():
    try:
        print(f"Client disconnected: {request.sid}")
        # Clean up subscriptions for this client
        for research_id, subscribers in list(socket_subscriptions.items()):
            if request.sid in subscribers:
                subscribers.remove(request.sid)
            if not subscribers:
                socket_subscriptions.pop(research_id, None)
                print(f"Removed empty subscription for research {research_id}")
    except Exception as e:
        print(f"Error handling disconnect: {e}")

@socketio.on('subscribe_to_research')
def handle_subscribe(data):
    research_id = data.get('research_id')
    if research_id:
        # First check if this research is still active
        conn = sqlite3.connect(DB_PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT status FROM research_history WHERE id = ?', (research_id,))
        result = cursor.fetchone()
        conn.close()
        
        # Only allow subscription to valid research
        if result:
            status = result[0]
            
            # Initialize subscription set if needed
            if research_id not in socket_subscriptions:
                socket_subscriptions[research_id] = set()
            
            # Add this client to the subscribers
            socket_subscriptions[research_id].add(request.sid)
            print(f"Client {request.sid} subscribed to research {research_id}")
            
            # Send current status immediately if available
            if research_id in active_research:
                progress = active_research[research_id]['progress']
                latest_log = active_research[research_id]['log'][-1] if active_research[research_id]['log'] else None
                
                if latest_log:
                    emit(f'research_progress_{research_id}', {
                        'progress': progress,
                        'message': latest_log.get('message', 'Processing...'),
                        'status': 'in_progress',
                        'log_entry': latest_log
                    })
            elif status in ['completed', 'failed', 'suspended']:
                # Send final status for completed research
                emit(f'research_progress_{research_id}', {
                    'progress': 100 if status == 'completed' else 0,
                    'message': 'Research completed successfully' if status == 'completed' else 
                               'Research failed' if status == 'failed' else 'Research was suspended',
                    'status': status,
                    'log_entry': {
                        'time': datetime.utcnow().isoformat(),
                        'message': f'Research is {status}',
                        'progress': 100 if status == 'completed' else 0,
                        'metadata': {'phase': 'complete' if status == 'completed' else 'error'}
                    }
                })
        else:
            # Research not found
            emit('error', {'message': f'Research ID {research_id} not found'})

@socketio.on_error
def handle_socket_error(e):
    print(f"Socket.IO error: {str(e)}")
    # Don't propagate exceptions to avoid crashing the server
    return False

@socketio.on_error_default
def handle_default_error(e):
    print(f"Unhandled Socket.IO error: {str(e)}")
    # Don't propagate exceptions to avoid crashing the server
    return False

# Function to clean up resources for a completed research
def cleanup_research_resources(research_id):
    """Clean up resources for a completed research"""
    print(f"Cleaning up resources for research {research_id}")
    
    # Get the current status from the database to determine the final status message
    current_status = "completed"  # Default
    try:
        conn = sqlite3.connect(DB_PATH)
        cursor = conn.cursor()
        cursor.execute('SELECT status FROM research_history WHERE id = ?', (research_id,))
        result = cursor.fetchone()
        if result and result[0]:
            current_status = result[0]
        conn.close()
    except Exception as e:
        print(f"Error retrieving research status during cleanup: {e}")
    
    # Remove from active research
    if research_id in active_research:
        del active_research[research_id]
        
    # Remove from termination flags
    if research_id in termination_flags:
        del termination_flags[research_id]
    
    # Send a final message to any remaining subscribers with explicit status
    if research_id in socket_subscriptions and socket_subscriptions[research_id]:
        # Use the proper status message based on database status
        if current_status == 'suspended' or current_status == 'failed':
            final_message = {
                'status': current_status,
                'message': f'Research was {current_status}',
                'progress': 0,  # For suspended research, show 0% not 100%
            }
        else:
            final_message = {
                'status': 'completed',
                'message': 'Research process has ended and resources have been cleaned up',
                'progress': 100,
            }
        
        try:
            print(f"Sending final {current_status} socket message for research {research_id}")
            # Use emit to all, not just subscribers
            socketio.emit(f'research_progress_{research_id}', final_message)
            
            # Also emit to specific subscribers
            for sid in socket_subscriptions[research_id]:
                try:
                    socketio.emit(
                        f'research_progress_{research_id}', 
                        final_message,
                        room=sid
                    )
                except Exception as sub_err:
                    print(f"Error emitting to subscriber {sid}: {str(sub_err)}")
        except Exception as e:
            print(f"Error sending final cleanup message: {e}")
    
    # Don't immediately remove subscriptions - let clients disconnect naturally

def run_research_process(research_id, query, mode):
    """Run the research process in the background for a given research ID"""
    try:
        # Check if this research has been terminated before we even start
        if research_id in termination_flags and termination_flags[research_id]:
            print(f"Research {research_id} was terminated before starting")
            cleanup_research_resources(research_id)
            return

        print(f"Starting research process for ID {research_id}, query: {query}")
        
        # Set up the AI Context Manager
        output_dir = os.path.join(OUTPUT_DIR, f"research_{research_id}")
        os.makedirs(output_dir, exist_ok=True)

        # Set up progress callback
        def progress_callback(message, progress_percent, metadata):
            # FREQUENT TERMINATION CHECK: Check for termination at each callback
            if research_id in termination_flags and termination_flags[research_id]:
                # Explicitly set the status to suspended in the database
                conn = sqlite3.connect(DB_PATH)
                cursor = conn.cursor()
                # Calculate duration up to termination point - using UTC consistently
                now = datetime.utcnow()
                completed_at = now.isoformat()
                
                # Get the start time from the database
                cursor.execute('SELECT created_at FROM research_history WHERE id = ?', (research_id,))
                result = cursor.fetchone()
                
                # Calculate the duration
                duration_seconds = calculate_duration(result[0]) if result and result[0] else None
                
                # Update the database with suspended status
                cursor.execute(
                    'UPDATE research_history SET status = ?, completed_at = ?, duration_seconds = ? WHERE id = ?',
                    ('suspended', completed_at, duration_seconds, research_id)
                )
                conn.commit()
                conn.close()
                
                # Clean up resources
                cleanup_research_resources(research_id)
                
                # Raise exception to exit the process
                raise Exception("Research was terminated by user")
            
            timestamp = datetime.utcnow().isoformat()
            
            # Adjust progress based on research mode
            adjusted_progress = progress_percent
            if mode == 'detailed' and metadata.get('phase') == 'output_generation':
                # For detailed mode, we need to adjust the progress range
                # because detailed reports take longer after the search phase
                adjusted_progress = min(80, progress_percent)
            elif mode == 'detailed' and metadata.get('phase') == 'report_generation':
                # Scale the progress from 80% to 95% for the report generation phase
                # Map progress_percent values (0-100%) to the (80-95%) range
                if progress_percent is not None:
                    normalized = progress_percent / 100
                    adjusted_progress = 80 + (normalized * 15)
            elif mode == 'quick' and metadata.get('phase') == 'output_generation':
                # For quick mode, ensure we're at least at 85% during output generation
                adjusted_progress = max(85, progress_percent)
                # Map any further progress within output_generation to 85-95% range
                if progress_percent is not None and progress_percent > 0:
                    normalized = progress_percent / 100
                    adjusted_progress = 85 + (normalized * 10)
            
            # Don't let progress go backwards
            if research_id in active_research and adjusted_progress is not None:
                current_progress = active_research[research_id].get('progress', 0)
                adjusted_progress = max(current_progress, adjusted_progress)
            
            log_entry = {
                "time": timestamp,
                "message": message,
                "progress": adjusted_progress,
                "metadata": metadata
            }
            
            # Check if termination was requested
            if research_id in termination_flags and termination_flags[research_id]:
                # Explicitly set the status to suspended in the database
                conn = sqlite3.connect(DB_PATH)
                cursor = conn.cursor()
                # Calculate duration up to termination point - using UTC consistently
                now = datetime.utcnow()
                completed_at = now.isoformat()
                
                # Get the start time from the database
                cursor.execute('SELECT created_at FROM research_history WHERE id = ?', (research_id,))
                result = cursor.fetchone()
                
                # Calculate the duration
                duration_seconds = calculate_duration(result[0]) if result and result[0] else None
                
                # Update the database with suspended status
                cursor.execute(
                    'UPDATE research_history SET status = ?, completed_at = ?, duration_seconds = ? WHERE id = ?',
                    ('suspended', completed_at, duration_seconds, research_id)
                )
                conn.commit()
                conn.close()
                
                # Clean up resources
                cleanup_research_resources(research_id)
                
                # Raise exception to exit the process
                raise Exception("Research was terminated by user")
            
            # Update active research record
            if research_id in active_research:
                active_research[research_id]['log'].append(log_entry)
                if adjusted_progress is not None:
                    active_research[research_id]['progress'] = adjusted_progress
                
                # Determine log type for database storage
                log_type = 'info'
                if metadata and metadata.get('phase'):
                    phase = metadata.get('phase')
                    if phase in ['complete', 'iteration_complete']:
                        log_type = 'milestone'
                    elif phase == 'error' or 'error' in message.lower():
                        log_type = 'error'
                
                # Always save logs to the new research_logs table
                add_log_to_db(
                    research_id,
                    message,
                    log_type=log_type,
                    progress=adjusted_progress,
                    metadata=metadata
                )
                
                # Update progress in the research_history table (for backward compatibility)
                conn = sqlite3.connect(DB_PATH)
                cursor = conn.cursor()
                
                # Update the progress and log separately to avoid race conditions with reading/writing the log
                if adjusted_progress is not None:
                    cursor.execute(
                        'UPDATE research_history SET progress = ? WHERE id = ?',
                        (adjusted_progress, research_id)
                    )
                
                # Add the log entry to the progress_log
                cursor.execute('SELECT progress_log FROM research_history WHERE id = ?', (research_id,))
                log_result = cursor.fetchone()
                
                if log_result:
                    try:
                        current_log = json.loads(log_result[0])
                    except:
                        current_log = []
                    
                    current_log.append(log_entry)
                    cursor.execute(
                        'UPDATE research_history SET progress_log = ? WHERE id = ?',
                        (json.dumps(current_log), research_id)
                    )
                
                conn.commit()
                conn.close()
                
                # Emit a socket event
                try:
                    # Basic event data
                    event_data = {
                        'message': message,
                        'progress': adjusted_progress
                    }
                    
                    # Add log entry in full format for detailed logging on client
                    if metadata:
                        event_data['log_entry'] = log_entry
                    
                    # Send to all subscribers and broadcast channel
                    socketio.emit(f'research_progress_{research_id}', event_data)
                    
                    if research_id in socket_subscriptions:
                        for sid in socket_subscriptions[research_id]:
                            try:
                                socketio.emit(
                                    f'research_progress_{research_id}', 
                                    event_data, 
                                    room=sid
                                )
                            except Exception as err:
                                print(f"Error emitting to subscriber {sid}: {str(err)}")
                except Exception as e:
                    print(f"Socket emit error (non-critical): {str(e)}")
                    
        # FUNCTION TO CHECK TERMINATION DURING LONG-RUNNING OPERATIONS
        def check_termination():
            if research_id in termination_flags and termination_flags[research_id]:
                # Explicitly set the status to suspended in the database
                conn = sqlite3.connect(DB_PATH)
                cursor = conn.cursor()
                now = datetime.utcnow()
                completed_at = now.isoformat()
                
                cursor.execute('SELECT created_at FROM research_history WHERE id = ?', (research_id,))
                result = cursor.fetchone()
                duration_seconds = calculate_duration(result[0]) if result and result[0] else None
                
                cursor.execute(
                    'UPDATE research_history SET status = ?, completed_at = ?, duration_seconds = ? WHERE id = ?',
                    ('suspended', completed_at, duration_seconds, research_id)
                )
                conn.commit()
                conn.close()
                
                # Clean up resources
                cleanup_research_resources(research_id)
                
                # Raise exception to exit the process
                raise Exception("Research was terminated by user during long-running operation")
            return False  # Not terminated

        # Set the progress callback in the system
        system = AdvancedSearchSystem()
        system.set_progress_callback(progress_callback)
        
        # Run the search
        progress_callback("Starting research process", 5, {"phase": "init"})
        
        try:
            results = system.analyze_topic(query)
            if mode == 'quick':
                progress_callback("Search complete, preparing to generate summary...", 85, {"phase": "output_generation"})
            else:
                progress_callback("Search complete, generating output", 80, {"phase": "output_generation"})
        except Exception as search_error:
            # Better handling of specific search errors
            error_message = str(search_error)
            error_type = "unknown"
            
            # Extract error details for common issues
            if "status code: 503" in error_message:
                error_message = "Ollama AI service is unavailable (HTTP 503). Please check that Ollama is running properly on your system."
                error_type = "ollama_unavailable"
            elif "status code: 404" in error_message:
                error_message = "Ollama model not found (HTTP 404). Please check that you have pulled the required model."
                error_type = "model_not_found"
            elif "status code:" in error_message:
                # Extract the status code for other HTTP errors
                status_code = error_message.split("status code:")[1].strip()
                error_message = f"API request failed with status code {status_code}. Please check your configuration."
                error_type = "api_error"
            elif "connection" in error_message.lower():
                error_message = "Connection error. Please check that your LLM service (Ollama/API) is running and accessible."
                error_type = "connection_error"
            
            # Raise with improved error message
            raise Exception(f"{error_message} (Error type: {error_type})")
        
        # Generate output based on mode
        if mode == 'quick':
            # Quick Summary
            if results.get('findings'):

                raw_formatted_findings = results['formatted_findings']
                logger.info(f"Found formatted_findings of length: {len(str(raw_formatted_findings))}")
                
                try:
                    clean_markdown = raw_formatted_findings
                    # ADDED CODE: Convert debug output to clean markdown
                    #clean_markdown = convert_debug_to_markdown(raw_formatted_findings, query)
                    print(f"Successfully converted to clean markdown of length: {len(clean_markdown)}")
                    
                    # First send a progress update for generating the summary
                    progress_callback("Generating clean summary from research data...", 90, {"phase": "output_generation"})
                    
                    # Save as markdown file
                    output_dir = "research_outputs"
                    if not os.path.exists(output_dir):
                        os.makedirs(output_dir)
                        
                    safe_query = "".join(x for x in query if x.isalnum() or x in [" ", "-", "_"])[:50]
                    safe_query = safe_query.replace(" ", "_").lower()
                    report_path = os.path.join(output_dir, f"quick_summary_{safe_query}.md")
                    
                    # Send progress update for writing to file
                    progress_callback("Writing research report to file...", 95, {"phase": "report_complete"})
                    
                    print(f"Writing report to: {report_path}")
                    with open(report_path, "w", encoding="utf-8") as f:
                        f.write("# Quick Research Summary\n\n")
                        f.write(f"Query: {query}\n\n")
                        f.write(clean_markdown)  # Use clean markdown instead of raw findings
                        f.write("\n\n## Research Metrics\n")
                        f.write(f"- Search Iterations: {results['iterations']}\n")
                        f.write(f"- Generated at: {datetime.utcnow().isoformat()}\n")
                    
                    # Update database
                    metadata = {
                        'iterations': results['iterations'],
                        'generated_at': datetime.utcnow().isoformat()
                    }
                    
                    # Calculate duration in seconds - using UTC consistently
                    now = datetime.utcnow()
                    completed_at = now.isoformat()
                    
                    print(f"Updating database for research_id: {research_id}")
                    # Get the start time from the database
                    conn = sqlite3.connect(DB_PATH)
                    cursor = conn.cursor()
                    cursor.execute('SELECT created_at FROM research_history WHERE id = ?', (research_id,))
                    result = cursor.fetchone()
                    
                    # Use the helper function for consistent duration calculation
                    duration_seconds = calculate_duration(result[0])
                    
                    # Update the record
                    cursor.execute(
                        'UPDATE research_history SET status = ?, completed_at = ?, duration_seconds = ?, report_path = ?, metadata = ? WHERE id = ?',
                        ('completed', completed_at, duration_seconds, report_path, json.dumps(metadata), research_id)
                    )
                    conn.commit()
                    conn.close()
                    print(f"Database updated successfully for research_id: {research_id}")
                    
                    # Send the final completion message
                    progress_callback("Research completed successfully", 100, {"phase": "complete", "report_path": report_path})
                    
                    # Clean up resources
                    print(f"Cleaning up resources for research_id: {research_id}")
                    cleanup_research_resources(research_id)
                    print(f"Resources cleaned up for research_id: {research_id}")
                except Exception as inner_e:
                    print(f"Error during quick summary generation: {str(inner_e)}")
                    print(traceback.format_exc())
                    raise Exception(f"Error generating quick summary: {str(inner_e)}")
            else:
                raise Exception("No research findings were generated. Please try again.")
        else:
            # Full Report
            progress_callback("Generating detailed report...", 85, {"phase": "report_generation"})
            report_generator = IntegratedReportGenerator()
            final_report = report_generator.generate_report(results, query)
            progress_callback("Report generation complete", 95, {"phase": "report_complete"})
            
            # Save as markdown file
            output_dir = "research_outputs"
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
                
            safe_query = "".join(x for x in query if x.isalnum() or x in [" ", "-", "_"])[:50]
            safe_query = safe_query.replace(" ", "_").lower()
            report_path = os.path.join(output_dir, f"detailed_report_{safe_query}.md")
            
            with open(report_path, "w", encoding="utf-8") as f:
                f.write(final_report['content'])
            
            # Update database
            metadata = final_report['metadata']
            metadata['iterations'] = results['iterations']
            
            # Calculate duration in seconds - using UTC consistently
            now = datetime.utcnow()
            completed_at = now.isoformat()
            
            # Get the start time from the database
            conn = sqlite3.connect(DB_PATH)
            cursor = conn.cursor()
            cursor.execute('SELECT created_at FROM research_history WHERE id = ?', (research_id,))
            result = cursor.fetchone()
            
            # Use the helper function for consistent duration calculation
            duration_seconds = calculate_duration(result[0])
            
            cursor.execute(
                'UPDATE research_history SET status = ?, completed_at = ?, duration_seconds = ?, report_path = ?, metadata = ? WHERE id = ?',
                ('completed', completed_at, duration_seconds, report_path, json.dumps(metadata), research_id)
            )
            conn.commit()
            conn.close()
            
            progress_callback("Research completed successfully", 100, {"phase": "complete", "report_path": report_path})
            
            # Clean up - moved to a separate function for reuse
            cleanup_research_resources(research_id)
            
    except Exception as e:
        # Handle error
        error_message = f"Research failed: {str(e)}"
        print(f"Research error: {error_message}")
        try:
            # Check for common Ollama error patterns in the exception and provide more user-friendly errors
            user_friendly_error = str(e)
            error_context = {}
            
            if "Error type: ollama_unavailable" in user_friendly_error:
                user_friendly_error = "Ollama AI service is unavailable. Please check that Ollama is running properly on your system."
                error_context = {"solution": "Start Ollama with 'ollama serve' or check if it's installed correctly."}
            elif "Error type: model_not_found" in user_friendly_error:
                user_friendly_error = "Required Ollama model not found. Please pull the model first."
                error_context = {"solution": "Run 'ollama pull mistral' to download the required model."}
            elif "Error type: connection_error" in user_friendly_error:
                user_friendly_error = "Connection error with LLM service. Please check that your AI service is running."
                error_context = {"solution": "Ensure Ollama or your API service is running and accessible."}
            elif "Error type: api_error" in user_friendly_error:
                # Keep the original error message as it's already improved
                error_context = {"solution": "Check API configuration and credentials."}
            
            # Update metadata with more context about the error
            metadata = {
                "phase": "error", 
                "error": user_friendly_error
            }
            if error_context:
                metadata.update(error_context)
                
            progress_callback(user_friendly_error, None, metadata)
        
            conn = sqlite3.connect(DB_PATH)
            cursor = conn.cursor()
            
            # If termination was requested, mark as suspended instead of failed
            status = 'suspended' if (research_id in termination_flags and termination_flags[research_id]) else 'failed'
            message = "Research was terminated by user" if status == 'suspended' else user_friendly_error
            
            # Calculate duration up to termination point - using UTC consistently
            now = datetime.utcnow()
            completed_at = now.isoformat()
            
            # Get the start time from the database
            duration_seconds = None
            cursor.execute('SELECT created_at FROM research_history WHERE id = ?', (research_id,))
            result = cursor.fetchone()
            
            # Use the helper function for consistent duration calculation
            if result and result[0]:
                duration_seconds = calculate_duration(result[0])
            
            cursor.execute(
                'UPDATE research_history SET status = ?, completed_at = ?, duration_seconds = ?, metadata = ? WHERE id = ?',
                (status, completed_at, duration_seconds, json.dumps(metadata), research_id)
            )
            conn.commit()
            conn.close()
            
            try:
                socketio.emit(f'research_progress_{research_id}', {
                    'status': status,
                    'error': message
                })
                
                # Also notify specific subscribers
                if research_id in socket_subscriptions and socket_subscriptions[research_id]:
                    for sid in socket_subscriptions[research_id]:
                        try:
                            socketio.emit(
                                f'research_progress_{research_id}', 
                                {'status': status, 'error': message},
                                room=sid
                            )
                        except Exception as sub_err:
                            print(f"Error emitting to subscriber {sid}: {str(sub_err)}")
                            
            except Exception as socket_error:
                print(f"Failed to emit error via socket: {str(socket_error)}")
        except Exception as inner_e:
            print(f"Error in error handler: {str(inner_e)}")
        
        # Clean up resources - moved to a separate function for reuse
        cleanup_research_resources(research_id)

@research_bp.route('/api/research/<int:research_id>/terminate', methods=['POST'])
def terminate_research(research_id):
    """Terminate an in-progress research process"""
    
    # Check if the research exists and is in progress
    conn = sqlite3.connect(DB_PATH)
    cursor = conn.cursor()
    cursor.execute('SELECT status FROM research_history WHERE id = ?', (research_id,))
    result = cursor.fetchone()
    
    if not result:
        conn.close()
        return jsonify({'status': 'error', 'message': 'Research not found'}), 404
    
    status = result[0]
    
    # If it's not in progress, return an error
    if status != 'in_progress':
        conn.close()
        return jsonify({'status': 'error', 'message': 'Research is not in progress'}), 400
    
    # Check if it's in the active_research dict
    if research_id not in active_research:
        # Update the status in the database
        cursor.execute('UPDATE research_history SET status = ? WHERE id = ?', ('suspended', research_id))
        conn.commit()
        conn.close()
        return jsonify({'status': 'success', 'message': 'Research terminated'})
    
    # Set the termination flag
    termination_flags[research_id] = True
    
    # Log the termination request - using UTC timestamp
    timestamp = datetime.utcnow().isoformat()
    termination_message = "Research termination requested by user"
    current_progress = active_research[research_id]['progress']
    
    # Create log entry
    log_entry = {
        "time": timestamp,
        "message": termination_message,
        "progress": current_progress,
        "metadata": {"phase": "termination"}
    }
    
    # Add to in-memory log
    active_research[research_id]['log'].append(log_entry)
    
    # Add to database log
    add_log_to_db(
        research_id,
        termination_message,
        log_type='milestone',
        progress=current_progress,
        metadata={"phase": "termination"}
    )
    
    # Update the log in the database (old way for backward compatibility)
    cursor.execute('SELECT progress_log FROM research_history WHERE id = ?', (research_id,))
    log_result = cursor.fetchone()
    if log_result:
        try:
            current_log = json.loads(log_result[0])
        except:
            current_log = []
        current_log.append(log_entry)
        cursor.execute(
            'UPDATE research_history SET progress_log = ? WHERE id = ?',
            (json.dumps(current_log), research_id)
        )
    
    # IMMEDIATELY update the status to 'suspended' to avoid race conditions
    cursor.execute('UPDATE research_history SET status = ? WHERE id = ?', ('suspended', research_id))
    conn.commit()
    conn.close()
    
    # Emit a socket event for the termination request
    try:
        event_data = {
            'status': 'suspended',  # Changed from 'terminating' to 'suspended'
            'message': 'Research was suspended by user request'
        }
        
        socketio.emit(f'research_progress_{research_id}', event_data)
        
        if research_id in socket_subscriptions and socket_subscriptions[research_id]:
            for sid in socket_subscriptions[research_id]:
                try:
                    socketio.emit(
                        f'research_progress_{research_id}', 
                        event_data,
                        room=sid
                    )
                except Exception as err:
                    print(f"Error emitting to subscriber {sid}: {str(err)}")
                    
    except Exception as socket_error:
        print(f"Socket emit error (non-critical): {str(socket_error)}")
    
    return jsonify({'status': 'success', 'message': 'Research termination requested'})

@research_bp.route('/api/research/<int:research_id>/delete', methods=['DELETE'])
def delete_research(research_id):
    """Delete a research record"""
    conn = sqlite3.connect(DB_PATH)
    cursor = conn.cursor()
    
    # First check if the research exists and is not in progress
    cursor.execute('SELECT status, report_path FROM research_history WHERE id = ?', (research_id,))
    result = cursor.fetchone()
    
    if not result:
        conn.close()
        return jsonify({'status': 'error', 'message': 'Research not found'}), 404
    
    status, report_path = result
    
    # Don't allow deleting research in progress
    if status == 'in_progress' and research_id in active_research:
        conn.close()
        return jsonify({
            'status': 'error', 
            'message': 'Cannot delete research that is in progress'
        }), 400
    
    # Delete report file if it exists
    if report_path and os.path.exists(report_path):
        try:
            os.remove(report_path)
        except Exception as e:
            print(f"Error removing report file: {str(e)}")
    
    # Delete the database record
    cursor.execute('DELETE FROM research_history WHERE id = ?', (research_id,))
    conn.commit()
    conn.close()
    
    return jsonify({'status': 'success'})
@research_bp.route('/settings', methods=['GET'])
def settings_page():
    """Main settings dashboard with links to specialized config pages"""
    return render_template('settings_dashboard.html')

@research_bp.route('/settings/main', methods=['GET'])
def main_config_page():
    """Edit main configuration with search parameters"""
    return render_template('main_config.html', main_file_path=MAIN_CONFIG_FILE)

@research_bp.route('/settings/llm', methods=['GET'])
def llm_config_page():
    """Edit LLM configuration using raw file editor"""
    return render_template('llm_config.html', llm_file_path=LLM_CONFIG_FILE)

@research_bp.route('/settings/collections', methods=['GET']) 
def collections_config_page():
    """Edit local collections configuration using raw file editor"""
    return render_template('collections_config.html', collections_file_path=LOCAL_COLLECTIONS_FILE)

@research_bp.route('/settings/api_keys', methods=['GET'])
def api_keys_config_page():
    """Edit API keys configuration"""
    # Get the secrets file path
    secrets_file = CONFIG_DIR / ".secrets.toml"
    
    return render_template('api_keys_config.html', secrets_file_path=secrets_file)
# Add to the imports section
from local_deep_research.config import SEARCH_ENGINES_FILE

# Add a new route for search engines configuration page
@research_bp.route('/settings/search_engines', methods=['GET'])
def search_engines_config_page():
    """Edit search engines configuration using raw file editor"""
    # Read the current config file
    raw_config = ""
    try:
        with open(SEARCH_ENGINES_FILE, 'r') as f:
            raw_config = f.read()
    except Exception as e:
        flash(f'Error reading search engines configuration: {str(e)}', 'error')
        raw_config = "# Error reading configuration file"
    
    # Get list of engine names for display
    engine_names = []
    try:
        from local_deep_research.web_search_engines.search_engines_config import SEARCH_ENGINES
        engine_names = list(SEARCH_ENGINES.keys())
        engine_names.sort()  # Alphabetical order
    except Exception as e:
        logger.error(f"Error getting engine names: {e}")
    
    return render_template('search_engines_config.html', 
                          search_engines_file_path=SEARCH_ENGINES_FILE,
                          raw_config=raw_config,
                          engine_names=engine_names)

# Add a route to save search engines configuration
@research_bp.route('/api/save_search_engines_config', methods=['POST'])
def save_search_engines_config():
    try:
        data = request.get_json()
        raw_config = data.get('raw_config', '')
        
        # Validate TOML syntax
        try:
            toml.loads(raw_config)
        except toml.TomlDecodeError as e:
            return jsonify({'success': False, 'error': f'TOML syntax error: {str(e)}'})
        
        # Ensure directory exists
        os.makedirs(os.path.dirname(SEARCH_ENGINES_FILE), exist_ok=True)
        
        # Create a backup first
        backup_path = f"{SEARCH_ENGINES_FILE}.bak"
        if os.path.exists(SEARCH_ENGINES_FILE):
            import shutil
            shutil.copy2(SEARCH_ENGINES_FILE, backup_path)
        
        # Write new config
        with open(SEARCH_ENGINES_FILE, 'w') as f:
            f.write(raw_config)
            
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


# API endpoint to save raw LLM config
@research_bp.route('/api/save_llm_config', methods=['POST'])
def save_llm_config():
    try:
        data = request.get_json()
        raw_config = data.get('raw_config', '')
        
        # Validate Python syntax
        try:
            compile(raw_config, '<string>', 'exec')
        except SyntaxError as e:
            return jsonify({'success': False, 'error': f'Syntax error: {str(e)}'})
        
        # Ensure directory exists
        os.makedirs(os.path.dirname(LLM_CONFIG_FILE), exist_ok=True)
        
        # Create a backup first
        backup_path = f"{LLM_CONFIG_FILE}.bak"
        if os.path.exists(LLM_CONFIG_FILE):
            import shutil
            shutil.copy2(LLM_CONFIG_FILE, backup_path)
        
        # Write new config
        with open(LLM_CONFIG_FILE, 'w') as f:
            f.write(raw_config)
            
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})

# API endpoint to save raw collections config
@research_bp.route('/api/save_collections_config', methods=['POST'])
def save_collections_config():
    try:
        data = request.get_json()
        raw_config = data.get('raw_config', '')
        
        # Validate TOML syntax
        try:
            toml.loads(raw_config)
        except toml.TomlDecodeError as e:
            return jsonify({'success': False, 'error': f'TOML syntax error: {str(e)}'})
        
        # Ensure directory exists
        os.makedirs(os.path.dirname(LOCAL_COLLECTIONS_FILE), exist_ok=True)
        
        # Create a backup first
        backup_path = f"{LOCAL_COLLECTIONS_FILE}.bak"
        if os.path.exists(LOCAL_COLLECTIONS_FILE):
            import shutil
            shutil.copy2(LOCAL_COLLECTIONS_FILE, backup_path)
        
        # Write new config
        with open(LOCAL_COLLECTIONS_FILE, 'w') as f:
            f.write(raw_config)
        
        # Also trigger a reload in the collections system
        try:
            load_local_collections(reload=True)
        except Exception as reload_error:
            return jsonify({'success': True, 'warning': f'Config saved, but error reloading: {str(reload_error)}'})
            
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})

# API endpoint to save raw main config 
@research_bp.route('/api/save_main_config', methods=['POST'])
def save_raw_main_config():
    try:
        data = request.get_json()
        raw_config = data.get('raw_config', '')
        
        # Validate TOML syntax
        try:
            toml.loads(raw_config)
        except toml.TomlDecodeError as e:
            return jsonify({'success': False, 'error': f'TOML syntax error: {str(e)}'})
        
        # Ensure directory exists
        os.makedirs(os.path.dirname(MAIN_CONFIG_FILE), exist_ok=True)
        
        # Create a backup first
        backup_path = f"{MAIN_CONFIG_FILE}.bak"
        if os.path.exists(MAIN_CONFIG_FILE):
            import shutil
            shutil.copy2(MAIN_CONFIG_FILE, backup_path)
        
        # Write new config
        with open(MAIN_CONFIG_FILE, 'w') as f:
            f.write(raw_config)
            
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})
@research_bp.route('/raw_config')
def get_raw_config():
    """Return the raw configuration file content"""
    try:
        # Determine which config file to load based on a query parameter
        config_type = request.args.get('type', 'main')
        
        if config_type == 'main':
            config_path = os.path.join(app.config['CONFIG_DIR'], 'config.toml')
            with open(config_path, 'r') as f:
                return f.read()
        elif config_type == 'llm':
            config_path = os.path.join(app.config['CONFIG_DIR'], 'llm_config.py')
            with open(config_path, 'r') as f:
                return f.read()
        elif config_type == 'collections':
            config_path = os.path.join(app.config['CONFIG_DIR'], 'collections.toml')
            with open(config_path, 'r') as f:
                return f.read()
        else:
            return "Unknown configuration type", 400
    except Exception as e:
        return str(e), 500
import os
import subprocess
import platform

@research_bp.route('/open_file_location', methods=['POST'])
def open_file_location():
    file_path = request.form.get('file_path')
    
    if not file_path:
        flash('No file path provided', 'error')
        return redirect(url_for('research.settings_page'))
    
    # Get the directory containing the file
    dir_path = os.path.dirname(os.path.abspath(file_path))
    
    # Open the directory in the file explorer
    try:
        if platform.system() == "Windows":
            subprocess.Popen(f'explorer "{dir_path}"')
        elif platform.system() == "Darwin":  # macOS
            subprocess.Popen(["open", dir_path])
        else:  # Linux
            subprocess.Popen(["xdg-open", dir_path])
        
        flash(f'Opening folder: {dir_path}', 'success')
    except Exception as e:
        flash(f'Error opening folder: {str(e)}', 'error')
    
    # Redirect back to the settings page
    if 'llm' in file_path:
        return redirect(url_for('research.llm_config_page'))
    elif 'collections' in file_path:
        return redirect(url_for('research.collections_config_page'))
    else:
        return redirect(url_for('research.main_config_page'))

@research_bp.route('/api/research/<int:research_id>/logs')
def get_research_logs(research_id):
    """Get logs for a specific research ID"""
    # First check if the research exists
    conn = sqlite3.connect(DB_PATH)
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()
    cursor.execute('SELECT id FROM research_history WHERE id = ?', (research_id,))
    result = cursor.fetchone()
    conn.close()
    
    if not result:
        return jsonify({'status': 'error', 'message': 'Research not found'}), 404
    
    # Retrieve logs from the database
    logs = get_logs_for_research(research_id)
    
    # Add any current logs from memory if this is an active research
    if research_id in active_research and active_research[research_id].get('log'):
        # Use the logs from memory temporarily until they're saved to the database
        memory_logs = active_research[research_id]['log']
        
        # Filter out logs that are already in the database
        # We'll compare timestamps to avoid duplicates
        db_timestamps = {log['time'] for log in logs}
        unique_memory_logs = [log for log in memory_logs if log['time'] not in db_timestamps]
        
        # Add unique memory logs to our return list
        logs.extend(unique_memory_logs)
        
        # Sort logs by timestamp
        logs.sort(key=lambda x: x['time'])
    
    return jsonify({
        'status': 'success',
        'logs': logs
    })



# Register the blueprint
app.register_blueprint(research_bp)

# Also add the static route at the app level for compatibility
@app.route('/static/<path:path>')
def app_serve_static(path):
    return send_from_directory(app.static_folder, path)

# Add favicon route to prevent 404 errors
@app.route('/favicon.ico')
def favicon():
    return send_from_directory(app.static_folder, 'favicon.ico', mimetype='image/x-icon')

# Add this function to app.py
def convert_debug_to_markdown(raw_text, query):
    """
    Convert the debug-formatted text to clean markdown.
    
    Args:
        raw_text: The raw formatted findings with debug symbols
        query: Original research query
    
    Returns:
        Clean markdown formatted text
    """
    try:
        print(f"Starting markdown conversion for query: {query}")
        print(f"Raw text type: {type(raw_text)}")
        
        # Handle None or empty input
        if not raw_text:
            print("WARNING: raw_text is empty or None")
            return f"No detailed findings available for '{query}'."
            
        # If there's a "DETAILED FINDINGS:" section, extract everything after it
        if "DETAILED FINDINGS:" in raw_text:
            print("Found DETAILED FINDINGS section")
            detailed_index = raw_text.index("DETAILED FINDINGS:")
            content = raw_text[detailed_index + len("DETAILED FINDINGS:"):].strip()
        else:
            print("No DETAILED FINDINGS section found, using full text")
            content = raw_text
        
        # Remove divider lines with === symbols
        lines_before = len(content.split("\n"))
        content = "\n".join([line for line in content.split("\n") 
                            if not line.strip().startswith("===") and not line.strip() == "="*80])
        lines_after = len(content.split("\n"))
        print(f"Removed {lines_before - lines_after} divider lines")
        

        
        # Remove SEARCH QUESTIONS BY ITERATION section
        if "SEARCH QUESTIONS BY ITERATION:" in content:
            print("Found SEARCH QUESTIONS BY ITERATION section")
            search_index = content.index("SEARCH QUESTIONS BY ITERATION:")
            next_major_section = -1
            for marker in ["DETAILED FINDINGS:", "COMPLETE RESEARCH:"]:
                if marker in content[search_index:]:
                    marker_pos = content.index(marker, search_index)
                    if next_major_section == -1 or marker_pos < next_major_section:
                        next_major_section = marker_pos
            
            if next_major_section != -1:
                print(f"Removing section from index {search_index} to {next_major_section}")
                content = content[:search_index] + content[next_major_section:]
            else:
                # If no later section, just remove everything from SEARCH QUESTIONS onwards
                print(f"Removing everything after index {search_index}")
                content = content[:search_index].strip()
        
        print(f"Final markdown length: {len(content.strip())}")
        return content.strip()
    except Exception as e:
        print(f"Error in convert_debug_to_markdown: {str(e)}")
        print(traceback.format_exc())
        # Return a basic message with the original query as fallback
        return f"# Research on {query}\n\nThere was an error formatting the research results."

def main():
    """
    Entry point for the web application when run as a command.
    This function is needed for the package's entry point to work properly.
    """
    # Import settings here to avoid circular imports
    from local_deep_research.config import settings

    # Get web server settings with defaults
    port = settings.web.port
    host = settings.web.host
    debug = settings.web.debug

    # Check for OpenAI availability but don't import it unless necessary
    try:
        import os
        api_key = os.environ.get("OPENAI_API_KEY")
        if api_key:
            try:
                # Only try to import if we have an API key
                import openai
                openai.api_key = api_key
                OPENAI_AVAILABLE = True
                print("OpenAI integration is available")
            except ImportError:
                print("OpenAI package not installed, integration disabled")
        else:
            print("OPENAI_API_KEY not found in environment variables, OpenAI integration disabled")
    except Exception as e:
        print(f"Error checking OpenAI availability: {e}")
        
    socketio.run(app, debug=debug, host=host, port=port, allow_unsafe_werkzeug=True)
    
if __name__ == '__main__':
    main()


import re


def remove_think_tags(text: str) -> str:
    text = re.sub(r"<think>.*?</think>", "", text, flags=re.DOTALL).strip()
    return text


def extract_links_from_search_results(search_results: list) -> list:
    """
    Extracts links and titles from a list of search result dictionaries.

    Each dictionary is expected to have at least the keys "title" and "link".

    Returns a list of dictionaries with 'title' and 'url' keys.
    """
    links = []
    for result in search_results:

        try:
            
            title = result.get("title", "").strip()
            url = result.get("link", "").strip()
            index = result.get("index", "").strip()
            
            if title and url:
                links.append({"title": title, "url": url, "index": index})
        except Exception:
            continue
    return links

def format_links(links):
    formatted_links =""
    formatted_links += "SOURCES:\n"
    for i, link in enumerate(links, 1):
        formatted_links += f"{link['index']}. {link['title']}\n   URL: {link['url']}\n"
    formatted_links += "\n"
    return formatted_links


def format_findings_to_text(findings_list, current_knowledge, questions_by_iteration):
    formatted_text = "COMPLETE RESEARCH OUTPUT \n\n"

    # Store the full current knowledge

    formatted_text += f"{current_knowledge}\n\n"
    formatted_text += "=" * 80 + "\n\n"

    # Store questions by iteration
    formatted_text += "SEARCH QUESTIONS BY ITERATION:\n"
    for iter_num, questions in questions_by_iteration.items():
        formatted_text += f"\nIteration {iter_num}:\n"
        for i, q in enumerate(questions, 1):
            formatted_text += f"{i}. {q}\n"
    formatted_text += "\n" + "=" * 80 + "\n\n"

    # Store detailed findings
    formatted_text += "DETAILED FINDINGS:\n\n"
    all_links = []  # To collect all sources

    for finding in findings_list:
        # Phase header
        formatted_text += f"{'='*80}\n"
        formatted_text += f"PHASE: {finding['phase']}\n"
        formatted_text += f"{'='*80}\n\n"

        # If this is a follow-up phase, show the corresponding question
        if finding["phase"].startswith("Follow-up"):
            iteration = int(finding["phase"].split(".")[0].split()[-1])
            question_index = int(finding["phase"].split(".")[-1]) - 1
            if iteration in questions_by_iteration and question_index < len(
                questions_by_iteration[iteration]
            ):
                formatted_text += f"SEARCH QUESTION:\n{questions_by_iteration[iteration][question_index]}\n\n"

        # Content
        formatted_text += f"CONTENT:\n{finding['content']}\n\n"

        # Search results if they exist
        if "search_results" in finding:
            # formatted_text += "SEARCH RESULTS:\n"
            # formatted_text += f"{finding['search_results']}\n\n"

            # Extract and format links for this finding
            links = extract_links_from_search_results(finding["search_results"])
            if links:
                formatted_text += "SOURCES USED IN THIS SECTION:\n"
                for i, link in enumerate(links, 1):
                    formatted_text += f"{i}. {link['title']}\n   URL: {link['url']}\n"
                formatted_text += "\n"
                all_links.extend(links)

        formatted_text += f"{'_'*80}\n\n"

    # Add summary of all sources at the end
    if all_links:
        formatted_text += "\nALL SOURCES USED IN RESEARCH:\n"
        formatted_text += "=" * 80 + "\n\n"
        seen_urls = set()  # To prevent duplicates
        for i, link in enumerate(all_links, 1):
            if link["url"] not in seen_urls:
                formatted_text += f"{i}. {link['title']}\n   URL: {link['url']}\n"
                seen_urls.add(link["url"])
        formatted_text += "\n" + "=" * 80 + "\n"

    return formatted_text

def print_search_results(search_results):
    formatted_text=""
    links = extract_links_from_search_results(search_results)
    if links:
        formatted_text=format_links(links=links)
    logger.info(formatted_text)




from langchain_community.utilities import DuckDuckGoSearchAPIWrapper
from typing import Dict, List, Any, Optional
from langchain_core.language_models import BaseLLM

from local_deep_research.web_search_engines.search_engine_base import BaseSearchEngine
from local_deep_research.web_search_engines.engines.full_search import FullSearchResults  # Import the FullSearchResults class
import logging
logger = logging.getLogger(__name__)

class DuckDuckGoSearchEngine(BaseSearchEngine):
    """DuckDuckGo search engine implementation with two-phase retrieval"""
    
    def __init__(self,
                max_results: int = 10,
                region: str = "us",
                safe_search: bool = True,
                llm: Optional[BaseLLM] = None,
                language: str = "English",
                include_full_content: bool = False,
                max_filtered_results=5):
        """
        Initialize the DuckDuckGo search engine.
        
        Args:
            max_results: Maximum number of search results
            region: Region code for search results
            safe_search: Whether to enable safe search
            llm: Language model for relevance filtering
            language: Language for content processing
            include_full_content: Whether to include full webpage content in results
        """
        # Initialize the BaseSearchEngine with LLM, max_filtered_results, and max_results
        super().__init__(llm=llm, max_filtered_results=max_filtered_results, max_results=max_results)
        self.region = region
        self.safe_search = safe_search
        self.language = language
        self.include_full_content = include_full_content
        
        # Initialize the DuckDuckGo wrapper
        self.engine = DuckDuckGoSearchAPIWrapper(
            region=region,
            max_results=max_results,
            safesearch="moderate" if safe_search else "off"
        )
        
        # Initialize FullSearchResults if full content is requested
        if include_full_content and llm:
            self.full_search = FullSearchResults(
                llm=llm,
                web_search=self.engine,
                language=language,
                max_results=max_results,
                region=region,
                time="y",
                safesearch="Moderate" if safe_search else "Off"
            )
    
    def run(self, query: str) -> List[Dict[str, Any]]:
        """
        Execute a search using DuckDuckGo with the two-phase approach.
        Respects config parameters:
        - SEARCH_SNIPPETS_ONLY: If True, only returns snippets without full content
        - SKIP_RELEVANCE_FILTER: If True, returns all results without filtering
        
        Args:
            query: The search query
            
        Returns:
            List of search results
        """
        logger.info("---Execute a search using DuckDuckGo---")
        
        # Implementation of the two-phase approach (from parent class)
        return super().run(query)
    
    def _get_previews(self, query: str) -> List[Dict[str, Any]]:
        """
        Get preview information (titles and snippets) for initial search results.
        
        Args:
            query: The search query
            
        Returns:
            List of preview dictionaries with 'id', 'title', and 'snippet' keys
        """
        try:
            # Get search results from DuckDuckGo
            results = self.engine.results(query, max_results=self.max_results)
            
            if not isinstance(results, list):
                return []
            
            # Process results to get previews
            previews = []
            for i, result in enumerate(results):
                preview = {
                    "id": result.get("link"),  # Use URL as ID for DDG
                    "title": result.get("title", ""),
                    "snippet": result.get("snippet", ""),
                    "link": result.get("link", "")
                }
                
                previews.append(preview)
            
            return previews
            
        except Exception as e:
            logger.error(f"Error getting DuckDuckGo previews: {e}")
            return []
    
    def _get_full_content(self, relevant_items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Get full content for the relevant items by using FullSearchResults.
        
        Args:
            relevant_items: List of relevant preview dictionaries
            
        Returns:
            List of result dictionaries with full content
        """
        # If we have FullSearchResults, use it to get full content
        if hasattr(self, 'full_search'):
            return self.full_search._get_full_content(relevant_items)
        
        # Otherwise, just return the relevant items without full content
        return relevant_items